@@ -1,3 +1,33 @@
+/* ex: set tabstop=8 expandtab: */
+/*
+  Copyright (c) 2006 - 2020
+  CLST  - Radboud University
+  ILK   - Tilburg University
+
+  This file is part of frog:
+
+  A Tagger-Lemmatizer-Morphological-Analyzer-Dependency-Parser for
+  several languages
+
+  frog is free software; you can redistribute it and/or modify
+  it under the terms of the GNU General Public License as published by
+  the Free Software Foundation; either version 3 of the License, or
+  (at your option) any later version.
+
+  frog is distributed in the hope that it will be useful,
+  but WITHOUT ANY WARRANTY; without even the implied warranty of
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+  GNU General Public License for more details.
+
+  You should have received a copy of the GNU General Public License
+  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+  For questions and suggestions, see:
+      https://github.com/LanguageMachines/frog/issues
+  or send mail to:
+      lamasoftware (at ) science.ru.nl
+
+*/
 #ifndef CKYPARSER_H
 #define CKYPARSER_H
 
@@ -8,6 +38,7 @@
 
 enum dirType { ROOT, LEFT, RIGHT, ERROR };
 
+/// \brief base class to hold parse constraints
 class Constraint {
  public:
   enum ConstraintType { Incoming, Dependency, Direction };
@@ -31,6 +62,7 @@ class Constraint {
 std::ostream& operator<<( std::ostream& os, const Constraint& c );
 std::ostream& operator<<( std::ostream& os, const Constraint* c );
 
+/// \brief specialization of Constraint to hold a relation
 class HasIncomingRel: public Constraint {
  public:
  HasIncomingRel( int i, const std::string& r, double w ):
@@ -43,6 +75,7 @@ class HasIncomingRel: public Constraint {
   std::string relType;
 };
 
+/// \brief specialization of Constraint to hold a dependency
 class HasDependency: public Constraint {
  public:
  HasDependency( int i, int h, const std::string& r, double w ):
@@ -57,6 +90,7 @@ class HasDependency: public Constraint {
   int headType;
 };
 
+/// \brief specialization of Constraint to hold the direction of a Dependency
 class DependencyDirection: public Constraint {
  public:
  DependencyDirection( int i, const std::string& d, double w ):
@@ -80,6 +114,7 @@ class DependencyDirection: public Constraint {
   dirType direction;
 };
 
+/// \brief structure to hold best fit so far
 class SubTree {
  public:
  SubTree( double score, int r, const std::string& label ):
@@ -98,11 +133,13 @@ class SubTree {
   std::string _edgeLabel;
 };
 
+/// \brief helper structure to hold a head and a relation
 struct parsrel {
   std::string deprel;
   int head;
 };
 
+/// \brief the class to hold left and right results
 class chart_rec {
  public:
   SubTree l_True;
@@ -111,6 +148,7 @@ class chart_rec {
   SubTree r_False;
 };
 
+/// \brief The class that can run the parser
 class CKYParser {
 public:
   CKYParser( size_t, const std::vector<const Constraint*>&, TiCC::LogStream* );
