@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -34,22 +34,21 @@
 
 #include "frog/tagger_base.h"
 
+/// \brief a specialization of Basetagger to tag IOB tags
 class IOBTagger: public BaseTagger {
  public:
-  explicit IOBTagger( TiCC::LogStream *l ): BaseTagger( l, "IOB" ){};
+  explicit IOBTagger( TiCC::LogStream *l, TiCC::LogStream *d =0 ):
+  BaseTagger( l, d, "IOB" ){};
   bool init( const TiCC::Configuration& );
-  void addDeclaration( folia::Document& ) const;
-  void Classify( const std::vector<folia::Word*>& );
-  void post_process( const std::vector<folia::Word*>& );
+  void add_declaration( folia::Document&, folia::processor * ) const;
+  void Classify( frog_data& );
+  void post_process( frog_data& );
+  void add_result( const frog_data& fd,
+		   const std::vector<folia::Word*>& wv ) const;
  private:
-  void addChunk( folia::ChunkingLayer *,
-		 const std::vector<folia::Word*>&,
-		 const std::vector<double>&,
-		 const std::string&,
-		 const std::string& );
-  void addIOBTags( const std::vector<folia::Word*>&,
-		   const std::vector<std::string>&,
-		   const std::vector<double>& );
+  void addTag( frog_record&,
+	       const std::string&,
+	       double );
 };
 
 #endif // IOB_TAGGER_MOD_H
