@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -38,11 +38,13 @@
 #include "ticcutils/Configuration.h"
 #include "ticcutils/Unicode.h"
 #include "libfolia/folia.h"
+#include "frog/FrogData.h"
 
+/// \brief a helper class for Mwu. Stores needed information.
 class mwuAna {
   friend std::ostream& operator<< (std::ostream&, const mwuAna& );
  public:
-  mwuAna( folia::Word *, const std::string&, const std::string& );
+  mwuAna( const std::string&, bool, size_t );
   virtual ~mwuAna() {};
 
   void merge( const mwuAna * );
@@ -52,32 +54,34 @@ class mwuAna {
   }
 
   bool isSpec(){ return spec; };
-  folia::EntitiesLayer *addEntity( const std::string&,
-				   const std::string&,
-				   folia::Sentence *,
-				   folia::EntitiesLayer * );
+
+  size_t mwu_start;
+  size_t mwu_end;
 
  protected:
     mwuAna(){};
     std::string word;
     bool spec;
-    std::vector<folia::Word *> fwords;
 };
 
 #define mymap2 std::multimap<std::string, std::vector<std::string> >
 
+/// \brief provide all functionality to detect MWU's
 class Mwu {
   friend std::ostream& operator<< (std::ostream&, const Mwu& );
  public:
-  explicit Mwu(TiCC::LogStream*);
+  explicit Mwu( TiCC::LogStream*, TiCC::LogStream* );
   ~Mwu();
   void reset();
   bool init( const TiCC::Configuration& );
-  void addDeclaration( folia::Document& ) const;
-  void Classify( const std::vector<folia::Word *>& );
-  void add( folia::Word * );
+  void add_provenance( folia::Document&, folia::processor * ) const;
+  void Classify( frog_data& );
+  void add( frog_record& );
+  void add_result( const frog_data&,
+		   const std::vector<folia::Word*>& ) const;
+  /// return the value for \e mwu_tagset. (set via Configuration)
   std::string getTagset() const { return mwu_tagset; };
-  std::vector<mwuAna*>& getAna(){ return mWords; };
+  std::string version() const { return _version; };
  private:
   bool readsettings( const std::string&, const std::string&);
   bool read_mwus( const std::string& );
@@ -86,8 +90,9 @@ class Mwu {
   std::string mwuFileName;
   std::vector<mwuAna*> mWords;
   mymap2 MWUs;
-  TiCC::LogStream *mwuLog;
-  std::string version;
+  TiCC::LogStream *errLog;
+  TiCC::LogStream *dbgLog;
+  std::string _version;
   std::string textclass;
   std::string mwu_tagset;
   std::string glue_tag;
