@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -31,10 +31,11 @@
 
 #include "frog/Frog-util.h"
 
+#include <cstring>
 #include <set>
 #include <string>
 #include <ostream>
-
+#include "ticcutils/SocketBasics.h"
 #include "config.h"
 
 #include <sys/types.h>
@@ -46,6 +47,13 @@
 using namespace std;
 
 string prefix( const string& path, const string& fn ){
+  /// add a full path to the filename \e fn
+  /*!
+    \param path the path to add
+    \param fn the filename to be prefixed
+    \return a new path
+    If \e fn already contains (relative) path information, it is left unchanged.
+   */
   if ( fn.find( "/" ) == string::npos && !path.empty() ){
     // only append prefix when it isn't empty AND
     // NO path is specified in the filename
@@ -55,12 +63,19 @@ string prefix( const string& path, const string& fn ){
 }
 
 #ifdef HAVE_DIRENT_H
-void getFileNames( const string& dirName,
-		   const string& ext,
-		   set<string>& fileNames ){
+set<string> getFileNames( const string& dirName,
+			   const string& ext ){
+  /// extract a (sorted) list of file-names matching an extension pattern
+  /*!
+    \param dirName the search directory
+    \param ext the file extension we search
+    \return a sorted list og file-names
+  */
+  set<string> result;
   DIR *dir = opendir( dirName.c_str() );
-  if ( !dir )
-    return;
+  if ( !dir ){
+    return result;
+  }
   else {
     struct stat sb;
     struct dirent *entry = readdir( dir );
@@ -71,26 +86,38 @@ void getFileNames( const string& dirName,
 	     filename.rfind( ext ) != string::npos ) {
 	  string fullName = dirName + "/" + filename;
 	  if ( stat( fullName.c_str(), &sb ) >= 0 ){
-            if ( (sb.st_mode & S_IFMT) == S_IFREG )
-	      fileNames.insert( entry->d_name );
+            if ( (sb.st_mode & S_IFMT) == S_IFREG ){
+	      result.insert( entry->d_name );
+	    }
 	  }
 	}
       }
       entry = readdir( dir );
     }
     closedir( dir );
+    return result;
   }
 }
-
 #endif
 
-string getTime() {
-  time_t Time;
-  time(&Time);
-  tm curtime;
-  localtime_r(&Time,&curtime);
-  char buf[256];
-  strftime( buf, 100, "%Y-%m-%dT%X", &curtime );
-  string res = buf;
-  return res;
+string check_server( const string& host,
+		     const string& port,
+		     const string& name ){
+  /// check if a certain host:port is available for us
+  /*!
+    \param host the host we check
+    \param port the port we want to access
+    \param name extra information (used in diagnostics only)
+    \return "" on succes or an error message on failure
+  */
+  string outline;
+  Sockets::ClientSocket client;
+  if ( !client.connect( host, port ) ){
+    outline = "cannot open connection, " + host + ":" + port;
+    if ( !name.empty() ){
+      outline += " for " + name + " module";
+    }
+    outline += "\nmessage: (" + client.getMessage() + ")" ;
+  }
+  return outline;
 }
