@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -35,6 +35,7 @@
 #include <unistd.h>
 #include <cstdlib>
 #include <string>
+#include <map>
 #include <iostream>
 #include <fstream>
 #include <algorithm>
@@ -42,22 +43,22 @@
 #include "config.h"
 #include "ticcutils/Configuration.h"
 #include "ticcutils/PrettyPrint.h"
+#include "ticcutils/SocketBasics.h"
+#include "ticcutils/json.hpp"
 #include "timbl/TimblAPI.h"
 #include "frog/Frog-util.h"
 #include "frog/csidp.h"
+#include "frog/Parser.h"
 
 using namespace std;
 
 using TiCC::operator<<;
+using namespace nlohmann;
 
-#define LOG *TiCC::Log(parseLog)
-
-TiCC::Timer prepareTimer;
-TiCC::Timer relsTimer;
-TiCC::Timer pairsTimer;
-TiCC::Timer dirTimer;
-TiCC::Timer csiTimer;
+#define LOG *TiCC::Log(errLog)
+#define DBG *TiCC::Dbg(dbgLog)
 
+/// structure to store parsing results
 struct parseData {
   void clear() { words.clear(); heads.clear(); mods.clear(); mwus.clear(); }
   vector<string> words;
@@ -67,6 +68,12 @@ struct parseData {
 };
 
 ostream& operator<<( ostream& os, const parseData& pd ){
+  /// output a parseData structure (debugging only)
+  /*!
+    \param os the output stream
+    \param pd the parseData to dump
+    \return the stream
+  */
   os << "pd words " << pd.words << endl;
   os << "pd heads " << pd.heads << endl;
   os << "pd mods " << pd.mods << endl;
@@ -81,6 +88,13 @@ ostream& operator<<( ostream& os, const parseData& pd ){
 }
 
 bool Parser::init( const TiCC::Configuration& configuration ){
+  /// initialize a Parser from the configuration
+  /*!
+    \param configuration the config to use
+    \return true on succes
+    extract all needed information from the configuration and setup the parser
+    by creating 3 Timbl instances
+  */
   filter = 0;
   string pairsFileName;
   string pairsOptions = "-a1 +D -G0 +vdb+di";
@@ -98,16 +112,16 @@ bool Parser::init( const TiCC::Configuration& configuration ){
     int level;
     if ( TiCC::stringTo<int>( val, level ) ){
       if ( level > 5 ){
-	parseLog->setlevel( LogLevel::LogDebug );
+	dbgLog->setlevel( LogLevel::LogDebug );
       }
     }
   }
   val = configuration.lookUp( "version", "parser" );
   if ( val.empty() ){
-    version = "1.0";
+    _version = "1.0";
   }
   else {
-    version = val;
+    _version = val;
   }
   val = configuration.lookUp( "set", "parser" );
   if ( val.empty() ){
@@ -131,8 +145,9 @@ bool Parser::init( const TiCC::Configuration& configuration ){
     MWU_tagset = val;
   }
   string charFile = configuration.lookUp( "char_filter_file", "tagger" );
-  if ( charFile.empty() )
+  if ( charFile.empty() ){
     charFile = configuration.lookUp( "char_filter_file" );
+  }
   if ( !charFile.empty() ){
     charFile = prefix( configuration.configDir(), charFile );
     filter = new TiCC::UniFilter();
@@ -151,41 +166,81 @@ bool Parser::init( const TiCC::Configuration& configuration ){
       problem = true;
     }
   }
-  val = configuration.lookUp( "pairsFile", "parser" );
+
+  val = configuration.lookUp( "host", "parser" );
   if ( !val.empty() ){
-    pairsFileName = prefix( cDir, val );
+    // so using an external parser server
+    _host = val;
+    val = configuration.lookUp( "port", "parser" );
+    if ( !val.empty() ){
+      _port = val;
+    }
+    else {
+      LOG << "missing port option" << endl;
+      problem = true;
+    }
+    val = configuration.lookUp( "pairs_base", "parser" );
+    if ( !val.empty() ){
+      _pairs_base = val;
+    }
+    else {
+      LOG << "missing pairs_base option" << endl;
+      problem = true;
+    }
+    val = configuration.lookUp( "dirs_base", "parser" );
+    if ( !val.empty() ){
+      _dirs_base = val;
+    }
+    else {
+      LOG << "missing dirs_base option" << endl;
+      problem = true;
+    }
+    val = configuration.lookUp( "rels_base", "parser" );
+    if ( !val.empty() ){
+      _rels_base = val;
+    }
+    else {
+      LOG << "missing rels_base option" << endl;
+      problem = true;
+    }
   }
   else {
-    LOG << "missing pairsFile option" << endl;
-    problem = true;
-  }
-  val = configuration.lookUp( "pairsOptions", "parser" );
-  if ( !val.empty() ){
-    pairsOptions = val;
-  }
-  val = configuration.lookUp( "dirFile", "parser" );
-  if ( !val.empty() ){
-    dirFileName = prefix( cDir, val );
-  }
-  else {
-    LOG << "missing dirFile option" << endl;
-    problem = true;
-  }
-  val = configuration.lookUp( "dirOptions", "parser" );
-  if ( !val.empty() ){
-    dirOptions = val;
-  }
-  val = configuration.lookUp( "relsFile", "parser" );
-  if ( !val.empty() ){
-    relsFileName = prefix( cDir, val );
-  }
-  else {
-    LOG << "missing relsFile option" << endl;
-    problem = true;
-  }
-  val = configuration.lookUp( "relsOptions", "parser" );
-  if ( !val.empty() ){
-    relsOptions = val;
+    val = configuration.lookUp( "pairsFile", "parser" );
+    if ( !val.empty() ){
+      pairsFileName = prefix( cDir, val );
+    }
+    else {
+      LOG << "missing pairsFile option" << endl;
+      problem = true;
+    }
+    val = configuration.lookUp( "pairsOptions", "parser" );
+    if ( !val.empty() ){
+      pairsOptions = val;
+    }
+    val = configuration.lookUp( "dirFile", "parser" );
+    if ( !val.empty() ){
+      dirFileName = prefix( cDir, val );
+    }
+    else {
+      LOG << "missing dirFile option" << endl;
+      problem = true;
+    }
+    val = configuration.lookUp( "dirOptions", "parser" );
+    if ( !val.empty() ){
+      dirOptions = val;
+    }
+    val = configuration.lookUp( "relsFile", "parser" );
+    if ( !val.empty() ){
+      relsFileName = prefix( cDir, val );
+    }
+    else {
+      LOG << "missing relsFile option" << endl;
+      problem = true;
+    }
+    val = configuration.lookUp( "relsOptions", "parser" );
+    if ( !val.empty() ){
+      relsOptions = val;
+    }
   }
   if ( problem ) {
     return false;
@@ -199,37 +254,51 @@ bool Parser::init( const TiCC::Configuration& configuration ){
     textclass = "current";
   }
   bool happy = true;
-  pairs = new Timbl::TimblAPI( pairsOptions );
-  if ( pairs->Valid() ){
-    LOG << "reading " <<  pairsFileName << endl;
-    happy = pairs->GetInstanceBase( pairsFileName );
-  }
-  else {
-    LOG << "creating Timbl for pairs failed:"
-		   << pairsOptions << endl;
-    happy = false;
-  }
-  if ( happy ){
-    dir = new Timbl::TimblAPI( dirOptions );
-    if ( dir->Valid() ){
-      LOG << "reading " <<  dirFileName << endl;
-      happy = dir->GetInstanceBase( dirFileName );
+  if ( _host.empty() ){
+    pairs = new Timbl::TimblAPI( pairsOptions );
+    if ( pairs->Valid() ){
+      LOG << "reading " <<  pairsFileName << endl;
+      happy = pairs->GetInstanceBase( pairsFileName );
     }
     else {
-      LOG << "creating Timbl for dir failed:" << dirOptions << endl;
+      LOG << "creating Timbl for pairs failed:"
+	  << pairsOptions << endl;
       happy = false;
     }
     if ( happy ){
-      rels = new Timbl::TimblAPI( relsOptions );
-      if ( rels->Valid() ){
-	LOG << "reading " <<  relsFileName << endl;
-	happy = rels->GetInstanceBase( relsFileName );
+      dir = new Timbl::TimblAPI( dirOptions );
+      if ( dir->Valid() ){
+	LOG << "reading " <<  dirFileName << endl;
+	happy = dir->GetInstanceBase( dirFileName );
       }
       else {
-	LOG << "creating Timbl for rels failed:"
-		       << relsOptions << endl;
+	LOG << "creating Timbl for dir failed:" << dirOptions << endl;
 	happy = false;
       }
+      if ( happy ){
+	rels = new Timbl::TimblAPI( relsOptions );
+	if ( rels->Valid() ){
+	  LOG << "reading " <<  relsFileName << endl;
+	  happy = rels->GetInstanceBase( relsFileName );
+	}
+	else {
+	  LOG << "creating Timbl for rels failed:"
+	      << relsOptions << endl;
+	  happy = false;
+	}
+      }
+    }
+  }
+  else {
+    string mess = check_server( _host, _port, "DependencyParser" );
+    if ( !mess.empty() ){
+      LOG << "FAILED to find a server for the Dependency parser:" << endl;
+      LOG << mess << endl;
+      LOG << "timblserver not running??" << endl;
+      happy = false;
+    }
+    else {
+      LOG << "using Parser Timbl's on " << _host << ":" << _port << endl;
     }
   }
   isInit = happy;
@@ -237,30 +306,18 @@ bool Parser::init( const TiCC::Configuration& configuration ){
 }
 
 Parser::~Parser(){
+  /// destructor
   delete rels;
   delete dir;
   delete pairs;
-  delete parseLog;
-  delete filter;
-}
-
-static vector<folia::Word *> lookup( folia::Word *word,
-				     const vector<folia::Entity*>& entities ){
-  vector<folia::Word*> vec;
-  for ( const auto& ent : entities ){
-    vec = ent->select<folia::Word>();
-    if ( !vec.empty() ){
-      if ( vec[0]->id() == word->id() ) {
-	// cerr << "found " << vec << endl;
-	return vec;
-      }
-    }
-  }
-  vec.clear();
-  return vec;
 }
 
 vector<string> Parser::createPairInstances( const parseData& pd ){
+  /// create a list of Instances for the 'pairs' Timbl
+  /*!
+    \param pd the parsedata structure with words, heads and modifiers
+    \return a list of string where every string is a Timbl test instance
+  */
   vector<string> instances;
   const vector<string>& words = pd.words;
   const vector<string>& heads = pd.heads;
@@ -405,6 +462,11 @@ vector<string> Parser::createPairInstances( const parseData& pd ){
 }
 
 vector<string> Parser::createDirInstances( const parseData& pd ){
+  /// create a list of Instances for the 'dir' Timbl
+  /*!
+    \param pd the parsedata structure with words, heads and modifiers
+    \return a list of string where every string is a Timbl test instance
+  */
   vector<string> d_instances;
   const vector<string>& words = pd.words;
   const vector<string>& heads = pd.heads;
@@ -617,6 +679,11 @@ vector<string> Parser::createDirInstances( const parseData& pd ){
 }
 
 vector<string> Parser::createRelInstances( const parseData& pd ){
+  /// create a list of Instances for the 'rel' Timbl
+  /*!
+    \param pd the parsedata structure with words, heads and modifiers
+    \return a list of string where every string is a Timbl test instance
+  */
   vector<string> r_instances;
   const vector<string>& words = pd.words;
   const vector<string>& heads = pd.heads;
@@ -798,187 +865,284 @@ vector<string> Parser::createRelInstances( const parseData& pd ){
   return r_instances;
 }
 
-
-void Parser::addDeclaration( folia::Document& doc ) const {
-  doc.declare( folia::AnnotationType::DEPENDENCY, dep_tagset,
-	       "annotator='frog-depparse-" + version
-	       + "', annotatortype='auto'");
+void Parser::add_provenance( folia::Document& doc, folia::processor *main ) const {
+  /// add provenance information to the FoLiA document
+  /*!
+    \param doc the foLiA document we are working on
+    \param main the main processor (presumably Frog) we want to add a new one to
+  */
+  string _label = "dep-parser";
+  if ( !main ){
+    throw logic_error( "Parser::add_provenance() without arguments." );
+  }
+  folia::KWargs args;
+  args["name"] = _label;
+  args["generate_id"] = "auto()";
+  args["version"] = _version;
+  args["begindatetime"] = "now()";
+  folia::processor *proc = doc.add_processor( args, main );
+  args.clear();
+  args["processor"] = proc->id();
+  doc.declare( folia::AnnotationType::DEPENDENCY, dep_tagset, args );
 }
 
-parseData Parser::prepareParse( const vector<folia::Word *>& fwords ){
-  parseData pd;
-  folia::Sentence *sent = 0;
-  vector<folia::Entity*> entities;
-#pragma omp critical (foliaupdate)
-  {
-    sent = fwords[0]->sentence();
-    entities = sent->select<folia::Entity>(MWU_tagset);
+void extract( const string& tv, string& head, string& mods ){
+  /// spit a (CGN-like) tag into a head and a modifier part
+  /*!
+    \param tv a CGN-like tag
+    \param head the extacted head
+    \param mods the modfiers, concatenated using the '|' symbol
+
+    Example: the tag WW(pv,tgw,met-t) will be split into a head 'WW' and a
+    mods string 'pv|tgw|met-t'
+   */
+  vector<string> v = TiCC::split_at_first_of( tv, "()" );
+  head = v[0];
+  mods.clear();
+  if ( v.size() > 1 ){
+    vector<string> mv = TiCC::split_at( v[1], "," );
+    mods = mv[0];
+    for ( size_t i=1; i < mv.size(); ++i ){
+      mods += "|" + mv[i];
+    }
   }
-  for ( size_t i=0; i < fwords.size(); ++i ){
-    folia::Word *word = fwords[i];
-    vector<folia::Word*> mwuv = lookup( word, entities );
-    if ( !mwuv.empty() ){
+  else {
+    mods = ""; // HACK should be "__", but then there are differences!
+  }                                                      //     ^
+}                                                        //     |
+                                                         //     |
+parseData Parser::prepareParse( frog_data& fd ){         //     |
+  /// setup the parser for  action
+  /*!
+    \param fd the frog_data structure with the needed information
+    \return a parseData structure for further processing
+  */
+  parseData pd;                                          //     |
+  for ( size_t i = 0; i < fd.units.size(); ++i ){        //     |
+    if ( fd.mwus.find( i ) == fd.mwus.end() ){           //     |
+      string head;                                       //     |
+      string mods;                                       //     |
+      extract( fd.units[i].tag, head, mods );            //     |
+      string word_s = fd.units[i].word;                  //     |
+      // the word may contain spaces, remove them all!          |
+      word_s.erase(remove_if(word_s.begin(), word_s.end(), ::isspace), word_s.end());
+      pd.words.push_back( word_s );                      //     |
+      pd.heads.push_back( head );                        //     |
+      if ( mods.empty() ){                               //    \/
+	// HACK: make this bug-to-bug compatible with older versions.
+	// But in fact this should also be done for the mwu's loop below!
+	// now sometimes empty mods get appended there.
+	// extract should return "__", I suppose  (see above)
+	mods = "__";
+      }
+      pd.mods.push_back( mods );
+    }
+    else {
       string multi_word;
-      string head;
-      string mod;
-      for ( const auto& mwu : mwuv ){
-	icu::UnicodeString tmp;
-#pragma omp critical (foliaupdate)
-	{
-	  tmp = mwu->text( textclass );
-	}
-	if ( filter )
+      string multi_head;
+      string multi_mods;
+      for ( size_t k = i; k <= fd.mwus[i]; ++k ){
+	icu::UnicodeString tmp = TiCC::UnicodeFromUTF8( fd.units[k].word );
+	if ( filter ){
 	  tmp = filter->filter( tmp );
+	}
 	string ms = TiCC::UnicodeToUTF8( tmp );
 	// the word may contain spaces, remove them all!
 	ms.erase(remove_if(ms.begin(), ms.end(), ::isspace), ms.end());
-	multi_word += ms;
-	folia::PosAnnotation *postag
-	  = mwu->annotation<folia::PosAnnotation>( POS_tagset );
-	head += postag->feat("head");
-	vector<folia::Feature*> feats = postag->select<folia::Feature>();
-	for ( const auto& feat : feats ){
-	  mod += feat->cls();
-	  if ( &feat != &feats.back() ){
-	    mod += "|";
-	  }
+	string head;
+	string mods;
+	extract( fd.units[k].tag, head, mods );
+	if ( k == i ){
+	  multi_word = ms;
+	  multi_head = head;
+	  multi_mods = mods;
 	}
-	if ( &mwu != &mwuv.back() ){
-	  multi_word += "_";
-	  head += "_";
-	  mod += "_";
+	else {
+	  multi_word += "_" + ms;
+	  multi_head += "_" + head;
+	  multi_mods += "_" + mods;
 	}
       }
       pd.words.push_back( multi_word );
-      pd.heads.push_back( head );
-      pd.mods.push_back( mod );
-      pd.mwus.push_back( mwuv );
-      i += mwuv.size()-1;
-    }
-    else {
-      icu::UnicodeString tmp;
-#pragma omp critical (foliaupdate)
-      {
-	tmp = word->text( textclass );
-      }
-      if ( filter )
-	tmp = filter->filter( tmp );
-      string ms = TiCC::UnicodeToUTF8( tmp );
-      // the word may contain spaces, remove them all!
-      ms.erase(remove_if(ms.begin(), ms.end(), ::isspace), ms.end());
-      pd.words.push_back( ms );
-      folia::PosAnnotation *postag
-	= word->annotation<folia::PosAnnotation>( POS_tagset );
-      string head = postag->feat("head");
-      pd.heads.push_back( head );
-      string mod;
-      vector<folia::Feature*> feats = postag->select<folia::Feature>();
-      if ( feats.empty() ){
-	mod = "__";
-      }
-      else {
-	for ( const auto& feat : feats ){
-	  mod += feat->cls();
-	  if ( &feat != &feats.back() ){
-	    mod += "|";
-	  }
-	}
-      }
-      pd.mods.push_back( mod );
-      vector<folia::Word*> vec;
-      vec.push_back(word);
-      pd.mwus.push_back( vec );
+      pd.heads.push_back( multi_head );
+      pd.mods.push_back( multi_mods );
+      i = fd.mwus[i];
     }
   }
   return pd;
 }
 
-void appendResult( const vector<folia::Word *>& words,
-		   parseData& pd,
-		   const string& tagset,
-		   const string& textclass,
-		   const vector<int>& nums,
-		   const vector<string>& roles ){
-  folia::Sentence *sent = 0;
-#pragma omp critical (foliaupdate)
-  {
-    sent = words[0]->sentence();
-  }
-  folia::DependenciesLayer *dl = 0;
-  folia::KWargs args;
-  args["generate_id"] = sent->id();
-  args["set"] = tagset;
-#pragma omp critical (foliaupdate)
-  {
-    dl = new folia::DependenciesLayer( args, sent->doc() );
-    sent->append( dl );
-  }
-  for ( size_t i=0; i < nums.size(); ++i ){
-    if ( nums[i] != 0 ){
-      folia::KWargs args;
-      args["generate_id"] = dl->id();
-      args["class"] = roles[i];
-      args["set"] = tagset;
-      if ( textclass != "current" ){
-	args["textclass"] = textclass;
-      }
-#pragma omp critical (foliaupdate)
-      {
-	folia::Dependency *d = new folia::Dependency( args, sent->doc() );
-	dl->append( d );
-	folia::Headspan *dh = new folia::Headspan();
-	for ( const auto& wrd : pd.mwus[nums[i]-1] ){
-	  dh->append( wrd );
-	}
-	d->append( dh );
-	folia::DependencyDependent *dd = new folia::DependencyDependent();
-	for ( const auto& it : pd.mwus[i] ){
-	  dd->append( it );
-	}
-	d->append( dd );
-      }
-    }
-  }
-}
 
-void appendParseResult( const vector<folia::Word *>& words,
-			parseData& pd,
-			const string& tagset,
-			const string& textclass,
-			const vector<parsrel>& res ){
-  vector<int> nums;
-  vector<string> roles;
-  for ( const auto& it : res ){
-    nums.push_back( it.head );
-    roles.push_back( it.deprel );
-  }
-  appendResult( words, pd, tagset, textclass, nums, roles );
-}
-
-void timbl( Timbl::TimblAPI* tim,
-	    const vector<string>& instances,
-	    vector<timbl_result>& results ){
-  results.clear();
+vector<timbl_result> timbl( Timbl::TimblAPI* tim,
+			    const vector<string>& instances ){
+  /// call a Timbl experiment with a list of instances
+  /*!
+    \param tim The Timbl to use
+    \param instances the instances to feed to the Timbl
+    \return a list of timbl_result structures with the result of processing
+    all instances
+   */
+  vector<timbl_result> result;
   for ( const auto& inst : instances ){
     const Timbl::ValueDistribution *db;
     const Timbl::TargetValue *tv = tim->Classify( inst, db );
-    results.push_back( timbl_result( tv->Name(), db->Confidence(tv), db ) );
+    result.push_back( timbl_result( tv->Name(), db->Confidence(tv), db ) );
+  }
+  return result;
+}
+
+vector<pair<string,double>> parse_vd( const string& ds ){
+  /// parse a ValueDistribution string into a vector of class/value pairs
+  /*!
+    \param ds a string representation of a Timbl ValueDistribution
+    \return a vector of string/double pairs. Each pair is one class + it's
+    (relative) count
+  */
+  vector<pair<string,double>> result;
+  vector<string> parts = TiCC::split_at_first_of( ds, "{,}" );
+  for ( const auto& p : parts ){
+    vector<string> sd = TiCC::split( p );
+    assert( sd.size() == 2 );
+    result.push_back( make_pair( sd[0], TiCC::stringTo<double>( sd[1] ) ) );
+  }
+  return result;
+}
+
+vector<timbl_result> Parser::timbl_server( const string& base,
+					   const vector<string>& instances ){
+  /// call a Timbl Server with a list of instances
+  /*!
+    \param base used to select the Timbl to use (should be configured correctly)
+    \param instances the instances to feed to the Timbl Server
+    \return a list of timbl_result structures with the result of processing
+    all instances
+   */
+  vector<timbl_result> result;
+  Sockets::ClientSocket client;
+  if ( !client.connect( _host, _port ) ){
+    LOG << "failed to open connection, " << _host
+	<< ":" << _port << endl
+	<< "Reason: " << client.getMessage() << endl;
+    exit( EXIT_FAILURE );
+  }
+  DBG << "calling " << base << " server" << endl;
+  string line;
+  client.read( line );
+  json response;
+  try {
+    response = json::parse( line );
+  }
+  catch ( const exception& e ){
+    LOG << "json parsing failed on '" << line << "':"
+	<< e.what() << endl;
+    abort();
+  }
+  DBG << "received json data:" << response.dump(2) << endl;
+
+  json out_json;
+  out_json["command"] = "base";
+  out_json["param"] = base;
+  string out_line = out_json.dump() + "\n";
+  DBG << "sending BASE json data:" << out_line << endl;
+  client.write( out_line );
+  client.read( line );
+  try {
+    response = json::parse( line );
+  }
+  catch ( const exception& e ){
+    LOG << "json parsing failed on '" << line << "':"
+	<< e.what() << endl;
+    abort();
+  }
+  DBG << "received json data:" << response.dump(2) << endl;
+
+  // create json query struct
+  json query;
+  query["command"] = "classify";
+  json arr = json::array();
+  for ( const auto& i : instances ){
+    arr.push_back( i );
+  }
+  query["params"] = arr;
+  DBG << "send json" << query.dump(2) << endl;
+  // send it to the server
+  line = query.dump() + "\n";
+  client.write( line );
+  // receive json
+  client.read( line );
+  DBG << "received line:" << line << "" << endl;
+  try {
+    response = json::parse( line );
+  }
+  catch ( const exception& e ){
+    LOG << "json parsing failed on '" << line << "':"
+	<< e.what() << endl;
+    LOG << "the request to the server was: '" <<  query.dump(2) << "'" << endl;
+    abort();
+  }
+  DBG << "received json data:" << response.dump(2) << endl;
+  if ( !response.is_array() ){
+    string cat = response["category"];
+    double conf = response.value("confidence",0.0);
+    vector<pair<string,double>> vd = parse_vd(response["distribution"]);
+    result.push_back( timbl_result( cat, conf, vd ) );
+  }
+  else {
+    for ( const auto& it : response.items() ){
+      string cat = it.value()["category"];
+      double conf = it.value().value("confidence",0.0);
+      vector<pair<string,double>> vd = parse_vd( it.value()["distribution"] );
+      result.push_back( timbl_result( cat, conf, vd ) );
+    }
+  }
+  return result;
+}
+
+void appendParseResult( frog_data& fd,
+			const vector<parsrel>& res ){
+  /// transfer the outcome of the parser back into the fog_data structure
+  /*!
+    \param fd the frog_data we want to add to
+    \param res the parser's results
+  */
+  vector<int> nums;
+  vector<string> roles;
+  for ( const auto& it : res ){
+    if ( it.head == 0 && it.deprel.empty() ){
+      continue;
+    }
+    nums.push_back( it.head );
+    roles.push_back( it.deprel );
+  }
+  // cerr << "NUMS=" << nums << endl;
+  // cerr << "roles=" << roles << endl;
+  for ( size_t i = 0; i < nums.size(); ++i ){
+    fd.mw_units[i].parse_index = nums[i];
+    fd.mw_units[i].parse_role = roles[i];
   }
 }
 
-void Parser::Parse( const vector<folia::Word*>& words,
-		    TimerBlock& timers ){
+void Parser::Parse( frog_data& fd, TimerBlock& timers ){
+  /// Run the Parser on 1 frog_data structure
+  /*!
+    \param fd the frog_data structure with our input
+    \param timers the TimerBlock for measuring what we wasting
+
+    This function will run 3 Timbl's in parallel to get its information
+    which is the handled to the 'real' parsing CSIDP process
+  */
   timers.parseTimer.start();
   if ( !isInit ){
     LOG << "Parser is not initialized! EXIT!" << endl;
     throw runtime_error( "Parser is not initialized!" );
   }
-  if ( words.empty() ){
-    LOG << "unable to parse an analisis without words" << endl;
+  if ( fd.empty() ){
+    LOG << "unable to parse an analysis without words" << endl;
     return;
   }
+
   timers.prepareTimer.start();
-  parseData pd = prepareParse( words );
+  parseData pd = prepareParse( fd );
   timers.prepareTimer.stop();
   vector<timbl_result> p_results;
   vector<timbl_result> d_results;
@@ -989,21 +1153,36 @@ void Parser::Parse( const vector<folia::Word*>& words,
       {
 	timers.pairsTimer.start();
 	vector<string> instances = createPairInstances( pd );
-	timbl( pairs, instances, p_results );
+	if ( _host.empty() ){
+	  p_results = timbl( pairs, instances );
+	}
+	else {
+	  p_results = timbl_server( _pairs_base, instances );
+	}
 	timers.pairsTimer.stop();
       }
 #pragma omp section
       {
 	timers.dirTimer.start();
 	vector<string> instances = createDirInstances( pd );
-	timbl( dir, instances, d_results );
+	if ( _host.empty() ){
+	  d_results = timbl( dir, instances );
+	}
+	else {
+	  d_results = timbl_server( _dirs_base, instances );
+	}
 	timers.dirTimer.stop();
       }
 #pragma omp section
       {
 	timers.relsTimer.start();
 	vector<string> instances = createRelInstances( pd );
-	timbl( rels, instances, r_results );
+	if ( _host.empty() ){
+	  r_results = timbl( rels, instances );
+	}
+	else {
+	  r_results = timbl_server( _rels_base, instances );
+	}
 	timers.relsTimer.stop();
       }
   }
@@ -1014,8 +1193,74 @@ void Parser::Parse( const vector<folia::Word*>& words,
 			       d_results,
 			       pd.words.size(),
 			       maxDepSpan,
-			       parseLog );
+			       dbgLog );
   timers.csiTimer.stop();
-  appendParseResult( words, pd, dep_tagset, textclass, res );
+  appendParseResult( fd, res );
   timers.parseTimer.stop();
 }
+
+ParserBase::~ParserBase(){
+  delete errLog;
+  delete dbgLog;
+  delete filter;
+}
+
+void ParserBase::add_result( const frog_data& fd,
+			     const vector<folia::Word*>& wv ) const {
+  /// add the parser's conclusiong to the FoLiA we are working on
+  /*!
+    \param fd the frog_data with the parse outcome allready inserted
+    \param wv the original folia::Word elements.
+
+    This will create a DependencyLayer at the Sentence level of the \e wv with
+    Dependency records for every dependency found
+  */
+  //  DBG << "Parser::add_result:" << endl << fd << endl;
+  folia::Sentence *s = wv[0]->sentence();
+  folia::KWargs args;
+  if ( !s->id().empty() ){
+    args["generate_id"] = s->id();
+  }
+  args["set"] = getTagset();
+  // if ( textclass != "current" ){
+  //   args["textclass"] = textclass;
+  // }
+  folia::DependenciesLayer *el = new folia::DependenciesLayer( args, s->doc() );
+  s->append( el );
+  for ( size_t pos=0; pos < fd.mw_units.size(); ++pos ){
+    //    DBG << "POS=" << pos << endl;
+    string cls = fd.mw_units[pos].parse_role;
+    int dep_id = fd.mw_units[pos].parse_index;
+    if ( cls != "ROOT" && dep_id != 0 ){
+      if ( !el->id().empty() ){
+	args["generate_id"] = el->id();
+      }
+      args["class"] = cls;
+      if ( textclass != "current" ){
+	args["textclass"] = textclass;
+      }
+      args["set"] = getTagset();
+      folia::Dependency *e = new folia::Dependency( args, s->doc() );
+      el->append( e );
+      args.clear();
+      // if ( textclass != "current" ){
+      // 	args["textclass"] = textclass;
+      // }
+      //      LOG << "wv.size=" << wv.size() << endl;
+      folia::Headspan *dh = new folia::Headspan( args );
+      // DBG << "mw_units:" << fd.mw_units[pos] << endl;
+      size_t head_index = fd.mw_units[pos].parse_index-1;
+      // DBG << "head_index=" << head_index << endl;
+      for ( auto const& i : fd.mw_units[head_index].parts ){
+	// DBG << "i=" << i << endl;
+	dh->append( wv[i] );
+      }
+      e->append( dh );
+      folia::DependencyDependent *dd = new folia::DependencyDependent( args );
+      for ( auto const& i : fd.mw_units[pos].parts ){
+	dd->append( wv[i] );
+      }
+      e->append( dd );
+    }
+  }
+}
