@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -37,13 +37,22 @@
 #include "timbl/TimblAPI.h"
 #include "ticcutils/LogStream.h"
 #include "ticcutils/Configuration.h"
+#include "ticcutils/SocketBasics.h"
+#include "ticcutils/json.hpp"
 #include "frog/Frog-util.h"
 
 using namespace std;
+using namespace nlohmann;
 
-#define LOG *TiCC::Log(mblemLog)
+#define LOG *TiCC::Log(errLog)
+#define DBG *TiCC::Log(dbgLog)
 
-Mblem::Mblem( TiCC::LogStream *logstream ):
+/// create a Timbl based lemmatizer
+/*!
+  \param errlog a LogStream for errors
+  \param dbglog a LogStream for debugging
+*/
+Mblem::Mblem( TiCC::LogStream *errlog, TiCC::LogStream *dbglog ):
   myLex(0),
   punctuation( "?...,:;\\'`(){}[]%#+-_=/!" ),
   history(20),
@@ -51,10 +60,35 @@ Mblem::Mblem( TiCC::LogStream *logstream ):
   keep_case( false ),
   filter(0)
 {
-  mblemLog = new TiCC::LogStream( logstream, "mblem" );
+  errLog = new TiCC::LogStream( errlog, "mblem-" );
+  if ( dbglog ){
+    dbgLog = new TiCC::LogStream( dbglog, "mblem" );
+  }
+  else {
+    dbgLog = errLog;
+  }
 }
 
 bool Mblem::fill_ts_map( const string& file ){
+  /// read 'token-strip' rules from a file
+  /*!
+    \param file name of the inputfile
+    \return true when no errors were detected
+
+    token-strip rules look like this:
+    \verbatim
+    SPEC(deeleigen)	QUOTE-SUFFIX	1
+    SPEC(deeleigen)	WORD-WITHSUFFIX	2
+    \endverbatim
+
+    What they express is this: When the tagger assigned a tag like
+    SPEC(deeleigen) AND the tokenizer assigned a class of QUOTE-SUFFIX, then
+    the lemma is given by stripping the last character. or last 2 for the second
+    rule.
+
+    This assures that the lemma for \e Jan's will be \e Jan and for \e Alex'
+    will be \e Alex
+   */
   ifstream is( file );
   if ( !is ){
     LOG << "Unable to open file: '" << file << "'" << endl;
@@ -62,8 +96,9 @@ bool Mblem::fill_ts_map( const string& file ){
   }
   string line;
   while ( getline( is, line ) ){
-    if ( line.empty() || line[0] == '#' )
+    if ( line.empty() || line[0] == '#' ){
       continue;
+    }
     vector<string> parts = TiCC::split( line );
     if ( parts.size() != 3 ){
       LOG << "invalid line in: '" << file << "' (expected 3 parts)" << endl;
@@ -71,16 +106,46 @@ bool Mblem::fill_ts_map( const string& file ){
     }
     token_strip_map[parts[0]].insert( make_pair( parts[1], TiCC::stringTo<int>( parts[2] ) ) );
   }
-  if ( debug ){
-    LOG << "read token strip rules from: '" << file << "'" << endl;
+  if ( debug > 1 ){
+    DBG << "read token strip rules from: '" << file << "'" << endl;
   }
   return true;
 }
 
 bool Mblem::init( const TiCC::Configuration& config ) {
+  /// initialize the lemmatizer using the config
+  /*!
+    \param config the Configuration to use
+    \return true when no problems are detected
+  */
   LOG << "Initiating lemmatizer..." << endl;
   debug = 0;
-  string val = config.lookUp( "debug", "mblem" );
+  string val = config.lookUp( "host", "mblem" );
+  if ( !val.empty() ){
+    // assume we must use a Timbl Server for the MBLEM
+    _host = val;
+    val = config.lookUp( "port", "mblem" );
+    if ( val.empty() ){
+      LOG << "missing 'port' settings for host= " << _host << endl;
+      return false;
+    }
+    _port = val;
+    val = config.lookUp( "base", "mblem" );
+    if ( val.empty() ){
+      LOG << "missing 'base' settings for host= " << _host << endl;
+      return false;
+    }
+    _base = val;
+  }
+  else {
+    val = config.lookUp( "base", "mblem" ) + ":"
+      + config.lookUp( "port", "mblem" );
+    if ( val.size() > 1 ){
+      LOG << "missing 'host' settings for base:port " << val << endl;
+      return false;
+    }
+  }
+  val = config.lookUp( "debug", "mblem" );
   if ( val.empty() ){
     val = config.lookUp( "debug" );
   }
@@ -89,24 +154,25 @@ bool Mblem::init( const TiCC::Configuration& config ) {
   }
   val = config.lookUp( "version", "mblem" );
   if ( val.empty() ){
-    version = "1.0";
+    _version = "1.0";
+  }
+  else {
+    _version = val;
   }
-  else
-    version = val;
   val = config.lookUp( "set", "mblem" );
   if ( val.empty() ){
     tagset = "http://ilk.uvt.nl/folia/sets/frog-mblem-nl";
   }
-  else
+  else {
     tagset = val;
-
+  }
   val = config.lookUp( "set", "tagger" );
   if ( val.empty() ){
     POS_tagset = "http://ilk.uvt.nl/folia/sets/frog-mbpos-cgn";
   }
-  else
+  else {
     POS_tagset = val;
-
+  }
   string transName = config.lookUp( "transFile", "mblem" );
   if ( !transName.empty() ){
     transName = prefix( config.configDir(), transName );
@@ -114,13 +180,15 @@ bool Mblem::init( const TiCC::Configuration& config ) {
     cerr << "skipping : " << transName << endl;
   }
   string treeName = config.lookUp( "treeFile", "mblem"  );
-  if ( treeName.empty() )
+  if ( treeName.empty() ){
     treeName = "mblem.tree";
+  }
   treeName = prefix( config.configDir(), treeName );
 
   string charFile = config.lookUp( "char_filter_file", "mblem" );
-  if ( charFile.empty() )
+  if ( charFile.empty() ){
     charFile = config.lookUp( "char_filter_file" );
+  }
   if ( !charFile.empty() ){
     charFile = prefix( config.configDir(), charFile );
     filter = new TiCC::UniFilter();
@@ -130,8 +198,9 @@ bool Mblem::init( const TiCC::Configuration& config ) {
   string tokenStripFile = config.lookUp( "token_strip_file", "mblem" );
   if ( !tokenStripFile.empty() ){
     tokenStripFile = prefix( config.configDir(), tokenStripFile );
-    if ( !fill_ts_map( tokenStripFile ) )
+    if ( !fill_ts_map( tokenStripFile ) ){
       return false;
+    }
   }
 
   string one_one_tagS = config.lookUp( "one_one_tags", "mblem" );
@@ -155,14 +224,30 @@ bool Mblem::init( const TiCC::Configuration& config ) {
     textclass = "current";
   }
 
-  string opts = config.lookUp( "timblOpts", "mblem" );
-  if ( opts.empty() )
-    opts = "-a1";
-  //make it silent
-  opts += " +vs -vf -F TABBED";
-  //Read in (igtree) data
-  myLex = new Timbl::TimblAPI(opts);
-  return myLex->GetInstanceBase(treeName);
+  if ( _host.empty() ){
+    string opts = config.lookUp( "timblOpts", "mblem" );
+    if ( opts.empty() ){
+      opts = "-a1";
+    }
+    // make it silent
+    opts += " +vs -vf -F TABBED";
+    //Read in (igtree) data
+    myLex = new Timbl::TimblAPI(opts);
+    return myLex->GetInstanceBase(treeName);
+  }
+  else {
+    string mess = check_server( _host, _port, "MBLEM" );
+    if ( !mess.empty() ){
+      LOG << "FAILED to find an MBLEM server:" << endl;
+      LOG << mess << endl;
+      LOG << "timblserver not running??" << endl;
+      return false;
+    }
+    else {
+      LOG << "using MBLEM Timbl on " << _host << ":" << _port << endl;
+      return true;
+    }
+  }
 }
 
 Mblem::~Mblem(){
@@ -170,12 +255,20 @@ Mblem::~Mblem(){
   delete filter;
   delete myLex;
   myLex = 0;
-  delete mblemLog;
+  if ( errLog != dbgLog ){
+    delete dbgLog;
+  }
+  delete errLog;
 }
 
 string Mblem::make_instance( const icu::UnicodeString& in ) {
-  if (debug) {
-    LOG << "making instance from: " << in << endl;
+  /// convert a Unicode string into an instance for Timbl
+  /*!
+    \param in the UnicodeString representing 1 word to lemmatize
+    \return an UTF8 string with an instance we can feed to Timbl
+  */
+  if (debug > 2 ) {
+    DBG << "making instance from: " << in << endl;
   }
   icu::UnicodeString instance = "";
   size_t length = in.length();
@@ -191,139 +284,131 @@ string Mblem::make_instance( const icu::UnicodeString& in ) {
   }
   instance += "?";
   string result = TiCC::UnicodeToUTF8(instance);
-  if ( debug ){
-    LOG << "inst: " << instance << endl;
+  if ( debug > 2 ){
+    DBG << "inst: " << instance << endl;
   }
   return result;
 }
 
-void Mblem::addLemma( folia::Word *word, const string& cls ){
-  folia::KWargs args;
-  args["set"]=tagset;
-  args["class"]=cls;
-  if ( textclass != "current" ){
-    args["textclass"] = textclass;
-  }
-
-#pragma omp critical (foliaupdate)
-  {
-    try {
-      word->addLemmaAnnotation( args );
-    }
-    catch( const exception& e ){
-      LOG << e.what() << " addLemma failed." << endl;
-      throw;
-    }
-  }
-}
-
 void Mblem::filterTag( const string& postag ){
+  /// filater all non-matching tags out of the mblem results
+  /*!
+    \param postag the tag, given by the CGN-tagger, that should match
+
+    Mblem produces a range of possible solutions with tags. We use the POS tag
+    given by the CGN tagger to remove all solutions with a different tag
+  */
   auto it = mblemResult.begin();
   while( it != mblemResult.end() ){
     string tag = it->getTag();
     if ( postag == tag ){
-      if ( debug ){
-	LOG << "compare cgn-tag " << postag << " with mblem-tag " << tag
-		       << "\n\t==> identical tags"  << endl;
+      if ( debug > 1 ){
+	DBG << "compare cgn-tag " << postag << " with mblem-tag " << tag
+	    << "\n\t==> identical tags"  << endl;
       }
       ++it;
     }
     else {
-      if ( debug ){
-	LOG << "compare cgn-tag " << postag << " with mblem-tag " << tag
+      if ( debug > 1 ){
+	DBG << "compare cgn-tag " << postag << " with mblem-tag " << tag
 		       << "\n\t==> different tags" << endl;
       }
       it = mblemResult.erase(it);
     }
   }
-  if ( debug && mblemResult.empty() ){
-    LOG << "NO CORRESPONDING TAG! " << postag << endl;
+  if ( (debug > 1) && mblemResult.empty() ){
+    DBG << "NO CORRESPONDING TAG! " << postag << endl;
   }
 }
 
 void Mblem::makeUnique( ){
+  /// filter out all results that are equal
+  /*
+    should be called AFTER filterTag() and cleans out doubles
+  */
   auto it = mblemResult.begin();
   while( it != mblemResult.end() ){
     string lemma = it->getLemma();
     auto it2 = it+1;
     while( it2 != mblemResult.end() ){
-      if (debug){
-	LOG << "compare lemma " << lemma << " with " << it2->getLemma() << " ";
+      if (debug > 1){
+	DBG << "compare lemma " << lemma << " with " << it2->getLemma() << " ";
       }
       if ( lemma == it2->getLemma() ){
-	if ( debug ){
-	  LOG << "equal " << endl;
+	if ( debug > 1){
+	  DBG << "equal " << endl;
 	}
 	it2 = mblemResult.erase(it2);
       }
       else {
-	if ( debug ){
-	  LOG << "NOT equal! " << endl;
+	if ( debug > 1){
+	  DBG << "NOT equal! " << endl;
 	}
 	++it2;
       }
     }
     ++it;
   }
-  if (debug){
-    LOG << "final result after filter and unique" << endl;
+  if (debug > 1){
+    DBG << "final result after filter and unique" << endl;
     for ( const auto& mbr : mblemResult ){
-      LOG << "lemma alt: " << mbr.getLemma()
+      DBG << "lemma alt: " << mbr.getLemma()
 	  << "\ttag alt: " << mbr.getTag() << endl;
     }
   }
 }
 
-void Mblem::getFoLiAResult( folia::Word *word, const icu::UnicodeString& uWord ){
-  if ( mblemResult.empty() ){
-    // just return the word as a lemma
-    string result = TiCC::UnicodeToUTF8( uWord );
-    addLemma( word, result );
-  }
-  else {
-    for ( auto const& it : mblemResult ){
-      string result = it.getLemma();
-      addLemma( word, result );
-    }
+void Mblem::add_provenance( folia::Document& doc,
+			    folia::processor *main ) const {
+  /// add provenance information to the FoLiA document
+  /*!
+    \param doc the foLiA document we are working on
+    \param main the main processor (presumably Frog) we want to add a new one to
+  */
+  string _label = "mblem";
+  if ( !main ){
+    throw logic_error( "mblem::add_provenance() without arguments." );
   }
+  folia::KWargs args;
+  args["name"] = _label;
+  args["generate_id"] = "auto()";
+  args["version"] = _version;
+  args["begindatetime"] = "now()";
+  folia::processor *proc = doc.add_processor( args, main );
+  args.clear();
+  args["processor"] = proc->id();
+  doc.declare( folia::AnnotationType::LEMMA, tagset, args );
 }
 
+void Mblem::Classify( frog_record& fd ){
+  /// add lemma information to the frog_data
+  /*!
+    \param fd The frog_data
 
-void Mblem::addDeclaration( folia::Document& doc ) const {
-  doc.declare( folia::AnnotationType::LEMMA,
-	       tagset,
-	       "annotator='frog-mblem-" + version
-	       + "', annotatortype='auto', datetime='" + getTime() + "'");
-}
-
-void Mblem::Classify( folia::Word *sword ){
-  if ( sword->isinstance( folia::PlaceHolder_t ) )
-    return;
+    this handles some special cases like ABBREVIATION, the token-strip rules
+    and the one-one rules.
+    All 'normal' cases are handled over to the Timbl classifier
+  */
   icu::UnicodeString uword;
   string pos;
   string token_class;
-#pragma omp critical (foliaupdate)
-  {
-    uword = sword->text( textclass );
-    pos = sword->pos();
-    string txtcls = sword->textclass();
-    if ( txtcls == textclass ){
-      // so only use the word class is the textclass of the word
-      // matches the wanted text
-      token_class = sword->cls();
-    }
-  }
-  if (debug){
-    LOG << "Classify " << uword << "(" << pos << ") ["
+  uword = TiCC::UnicodeFromUTF8(fd.word);
+  pos = fd.tag;
+  token_class = fd.token_class;
+  if (debug > 1 ){
+    DBG << "Classify " << uword << "(" << pos << ") ["
 	<< token_class << "]" << endl;
   }
-  if ( filter )
+  if ( filter ){
     uword = filter->filter( uword );
-
+  }
   if ( token_class == "ABBREVIATION" ){
     // We dont handle ABBREVIATION's so just take the word as such
     string word = TiCC::UnicodeToUTF8(uword);
-    addLemma( sword, word );
+#pragma omp critical (dataupdate)
+    {
+      fd.lemmas.push_back( word );
+    }
     return;
   }
   auto const& it1 = token_strip_map.find( pos );
@@ -337,14 +422,20 @@ void Mblem::Classify( folia::Word *sword ){
 	uword2 = uword;
       }
       string word = TiCC::UnicodeToUTF8(uword2);
-      addLemma( sword, word );
+#pragma omp critical (dataupdate)
+      {
+	fd.lemmas.push_back( word );
+      }
       return;
     }
   }
   if ( one_one_tags.find(pos) != one_one_tags.end() ){
     // some tags are just taken as such
     string word = TiCC::UnicodeToUTF8(uword);
-    addLemma( sword, word );
+#pragma omp critical (dataupdate)
+    {
+      fd.lemmas.push_back( word );
+    }
     return;
   }
   if ( !keep_case ){
@@ -353,16 +444,115 @@ void Mblem::Classify( folia::Word *sword ){
   Classify( uword );
   filterTag( pos );
   makeUnique();
-  getFoLiAResult( sword, uword );
+  if ( mblemResult.empty() ){
+    // just return the word as a lemma
+    string word = TiCC::UnicodeToUTF8(uword);
+#pragma omp critical (dataupdate)
+    {
+      fd.lemmas.push_back( word );
+    }
+  }
+  else {
+#pragma omp critical (dataupdate)
+    {
+      for ( auto const& it : mblemResult ){
+	string result = it.getLemma();
+	fd.lemmas.push_back( result );
+      }
+    }
+  }
+}
+
+string Mblem::call_server( const string& instance ){
+  /// use a Timbl server to classify
+  /*!
+    \param instance The instance to give to Timbl
+    \return the lemma rule as returned by Timbl
+  */
+  Sockets::ClientSocket client;
+  if ( !client.connect( _host, _port ) ){
+    LOG << "failed to open connection, " << _host
+	<< ":" << _port << endl
+	<< "Reason: " << client.getMessage() << endl;
+    exit( EXIT_FAILURE );
+  }
+  if ( debug > 1 ){
+    DBG << "calling MBLEM-server" << endl;
+  }
+  string line;
+  client.read( line );
+  json response;
+  try {
+    response = json::parse( line );
+  }
+  catch ( const exception& e ){
+    LOG << "json parsing failed on '" << line << "':"
+	<< e.what() << endl;
+    abort();
+  }
+  //  LOG << "received json data:" << response.dump(2) << endl;
+  if ( !_base.empty() ){
+    json out_json;
+    out_json["command"] = "base";
+    out_json["param"] = _base;
+    string line = out_json.dump() + "\n";
+    //    LOG << "sending BASE json data:" << line << endl;
+    client.write( line );
+    client.read( line );
+    json response;
+    try {
+      response = json::parse( line );
+    }
+    catch ( const exception& e ){
+      LOG << "json parsing failed on '" << line << "':"
+	  << e.what() << endl;
+      abort();
+    }
+    //    LOG << "received json data:" << response.dump(2) << endl;
+  }
+  // create json struct
+  json query;
+  query["command"] = "classify";
+  query["param"] = instance;
+  //  LOG << "send json" << query.dump(2) << endl;
+  // send it to the server
+  line = query.dump() + "\n";
+  client.write( line );
+  // receive json
+  client.read( line );
+  //  LOG << "received line:" << line << "" << endl;
+  try {
+    response = json::parse( line );
+  }
+  catch ( const exception& e ){
+    LOG << "json parsing failed on '" << line << "':"
+	<< e.what() << endl;
+    abort();
+  }
+  //  LOG << "received json data:" << response.dump(2) << endl;
+  string result = response["category"];
+  //  LOG << "extracted result " << result << endl;
+  return result;
 }
 
 void Mblem::Classify( const icu::UnicodeString& uWord ){
+  /// give the lemma for 1 word
+  /*!
+    \param uWord a Unicode string with the word
+    the mblemResult struct will be filled with 1 or more (alternative) solutions
+    of a lemma + a POS-tag
+  */
   mblemResult.clear();
   string inst = make_instance(uWord);
   string classString;
-  myLex->Classify( inst, classString );
-  if ( debug){
-    LOG << "class: " << classString  << endl;
+  if ( !_host.empty() ){
+    classString = call_server( inst );
+  }
+  else {
+    myLex->Classify( inst, classString );
+  }
+  if ( debug > 1){
+    DBG << "class: " << classString  << endl;
   }
   // 1st find all alternatives
   vector<string> parts;
@@ -411,9 +601,9 @@ void Mblem::Classify( const icu::UnicodeString& uWord ){
 			 << endl;
 	}
       }
-      if (debug){
-	LOG << "pre-prefix word: '" << uWord << "' prefix: '"
-		       << prefix << "'" << endl;
+      if (debug > 1){
+	DBG << "pre-prefix word: '" << uWord << "' prefix: '"
+	    << prefix << "'" << endl;
       }
       int prefixpos = 0;
       if ( !prefix.isEmpty() ) {
@@ -421,8 +611,8 @@ void Mblem::Classify( const icu::UnicodeString& uWord ){
 	// 'tegemoetgekomen' example), this should probably
 	// become prefixpos = uWord.lastIndexOf(prefix);
 	prefixpos = uWord.indexOf(prefix);
-	if (debug){
-	  LOG << "prefixpos = " << prefixpos << endl;
+	if (debug > 1){
+	  DBG << "prefixpos = " << prefixpos << endl;
 	}
 	// repair cases where there's actually not a prefix present
 	if (prefixpos > uWord.length()-2) {
@@ -431,15 +621,15 @@ void Mblem::Classify( const icu::UnicodeString& uWord ){
 	}
       }
 
-      if (debug){
-	LOG << "prefixpos = " << prefixpos << endl;
+      if (debug > 1){
+	DBG << "prefixpos = " << prefixpos << endl;
       }
       if (prefixpos >= 0) {
 	lemma = icu::UnicodeString( uWord, 0L, prefixpos );
 	prefixpos = prefixpos + prefix.length();
       }
-      if (debug){
-	LOG << "post word: "<< uWord
+      if (debug > 1){
+	DBG << "post word: "<< uWord
 	    << " lemma: " << lemma
 	    << " prefix: " << prefix
 	    << " delstr: " << delstr
@@ -474,13 +664,13 @@ void Mblem::Classify( const icu::UnicodeString& uWord ){
     if ( !classMap.empty() ){
       // translate TAG(number) stuf back to CGN things
       auto const& it = classMap.find(restag);
-      if ( debug ){
-	LOG << "looking up " << restag << endl;
+      if ( debug > 1 ){
+	DBG << "looking up " << restag << endl;
       }
       if ( it != classMap.end() ){
 	restag = it->second;
-	if ( debug ){
-	  LOG << "found " << restag << endl;
+	if ( debug > 1 ){
+	  DBG << "found " << restag << endl;
 	}
       }
       else {
@@ -488,22 +678,23 @@ void Mblem::Classify( const icu::UnicodeString& uWord ){
 	    << restag << " using it 'as-is'" << endl;
       }
     }
-    if ( debug ){
-      LOG << "appending lemma " << lemma << " and tag " << restag << endl;
+    if ( debug > 1 ){
+      DBG << "appending lemma " << lemma << " and tag " << restag << endl;
     }
     mblemResult.push_back( mblemData( TiCC::UnicodeToUTF8(lemma), restag ) );
   } // while
-  if ( debug ) {
-    LOG << "stored lemma and tag options: " << mblemResult.size()
+  if ( debug > 1) {
+    DBG << "stored lemma and tag options: " << mblemResult.size()
 	<< " lemma's and " << mblemResult.size() << " tags:" << endl;
     for ( const auto& mbr : mblemResult ){
-      LOG << "lemma alt: " << mbr.getLemma()
+      DBG << "lemma alt: " << mbr.getLemma()
 	  << "\ttag alt: " << mbr.getTag() << endl;
     }
   }
 }
 
 vector<pair<string,string> > Mblem::getResult() const {
+  /// extract the results into a list of lemma/tag pairs
   vector<pair<string,string> > result;
   for ( const auto& mbr : mblemResult ){
     result.push_back( make_pair( mbr.getLemma(),
@@ -511,3 +702,27 @@ vector<pair<string,string> > Mblem::getResult() const {
   }
   return result;
 }
+
+void Mblem::add_lemmas( const vector<folia::Word*>& wv,
+			const frog_data& fd ) const {
+  /// add the lemma from 'fd' to the FoLiA list of Word
+  /*!
+    \param wv The folia:Word vector
+    \param fd the folia_data with added lemmatizer results
+  */
+  ///
+  for ( size_t i=0; i < wv.size(); ++i ){
+    folia::KWargs args;
+    args["set"] = getTagset();
+    for ( const auto& lemma : fd.units[i].lemmas ){
+      args["class"] = lemma;
+      if ( textclass != "current" ){
+	args["textclass"] = textclass;
+      }
+#pragma omp critical (foliaupdate)
+      {
+	wv[i]->addLemmaAnnotation( args );
+      }
+    }
+  }
+}
