@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -46,15 +46,19 @@
 #include "frog/cgn_tagger_mod.h"
 
 using namespace std;
+using namespace icu;
 using namespace Timbl;
 using namespace Tagger;
+using TiCC::operator<<;
 
 TiCC::LogStream my_default_log( cerr, "", StampMessage ); // fall-back
 TiCC::LogStream *theErrLog = &my_default_log;  // fill the externals
 
 vector<string> fileNames;
+bool usewordlist = false;
 bool useTagger = true;
 bool useTokenizer = true;
+string output_name;
 
 TiCC::Configuration configuration;
 static string configDir = string(SYSCONF_PATH) + "/" + PACKAGE + "/nld/";
@@ -68,6 +72,9 @@ void usage( ) {
        << "Options:n" << endl;
   cout << "\t============= INPUT MODE (mandatory, choose one) ========================\n"
        << "\t -t <testfile>    Run mblem on this file\n"
+       << "\t --wordlist <wordfile>    Run on a wordlist. Produces a lemma list\n"
+       << "\t -o <outputfile>    write results in 'outputfile'\n"
+
        << "\t --notokenizer    Don't use a tokenizer, so assume all text is tokenized already.\n"
        << "\t --notagger       Don't use a tagger to disambiguate, so give ALL variants.\n"
        << "\t -c <filename>    Set configuration file (default " << configFileName << ")\n"
@@ -112,6 +119,21 @@ bool parse_args( TiCC::CL_Options& Opts ) {
     configuration.setatt( "debug", value, "mblem" );
   };
 
+  if ( Opts.extract( 'o', value ) ){
+    ofstream os( value );
+    if ( !os ){
+      cerr << "output stream " << value << " is not writeable" << endl;
+      return false;
+    }
+    output_name = value;
+  }
+  useTagger = !Opts.is_present( "notagger" );
+  useTokenizer = !Opts.is_present( "notokenizer" );
+  usewordlist = Opts.is_present( "wordlist" );
+  if ( usewordlist ){
+    useTagger = false;
+    useTokenizer = false;
+  }
   if ( Opts.extract( 't', value ) ){
     ifstream is( value );
     if ( !is ){
@@ -123,8 +145,6 @@ bool parse_args( TiCC::CL_Options& Opts ) {
   else {
     fileNames = Opts.getMassOpts();
   };
-  useTagger = !Opts.is_present( "notagger" );
-  useTokenizer = !Opts.is_present( "notokenizer" );
   return true;
 }
 
@@ -152,9 +172,13 @@ bool init(){
   return true;
 }
 
-void Test( istream& in ){
+void Test( istream& in, ostream& os ){
   string line;
   while ( getline( in, line ) ){
+    if ( line.empty() ) {
+      os << endl;
+      continue;
+    }
     vector<string> sentences;
     if ( useTokenizer ){
       sentences = tokenizer.tokenize( line );
@@ -166,48 +190,84 @@ void Test( istream& in ){
       if ( useTagger ){
 	vector<TagResult> tagrv = tagger.tagLine( s );
 	for ( const auto& tr : tagrv ){
-	  icu::UnicodeString uWord = TiCC::UnicodeFromUTF8(tr.word());
+	  UnicodeString uWord = TiCC::UnicodeFromUTF8(tr.word());
 	  myMblem.Classify( uWord );
-	  myMblem.filterTag( tr.assignedTag() );
+	  myMblem.filterTag( tr.assigned_tag() );
 	  vector<pair<string,string> > res = myMblem.getResult();
-	  string line = tr.word() + " {" + tr.assignedTag() + "}\t";
+	  string line = tr.word() + " {" + tr.assigned_tag() + "}\t";
 	  for ( const auto& p : res ){
 	    line += p.first + "[" + p.second + "]/";
 	  }
 	  line.erase(line.length()-1);
-	  cout << line << endl;
+	  line += "\n";
+	  os << line;
 	}
       }
       else {
-	vector<string> parts = TiCC::split( s );
-	for ( const auto& w : parts ){
-	  icu::UnicodeString uWord = TiCC::UnicodeFromUTF8(w);
-	  myMblem.Classify( uWord );
-	  vector<pair<string,string> > res = myMblem.getResult();
-	  string line = w + "\t";
-	  for ( const auto& p : res ){
-	    line += p.first + "[" + p.second + "]/";
+	string line = s;
+	if ( usewordlist ){
+	  if ( line[0] == '"' ){
+	    line.erase(line.begin());
+	  }
+	  if ( line.back() == '"' ){
+	    line.pop_back();
+	  }
+	  vector<string> parts = TiCC::split( line );
+	  if ( parts.size() > 1 ){
+	    cerr << "skipping multiword " << s << endl;
+	  }
+	  else {
+	    UnicodeString uWord = TiCC::UnicodeFromUTF8(parts[0]);
+	    myMblem.Classify( uWord );
+	    vector<pair<string,string> > res = myMblem.getResult();
+	    string line =  parts[0] + ",";
+	    set<string> out_set;
+	    for ( const auto& p : res ){
+	      if ( p.first != parts[0] || out_set.empty() ){
+		out_set.insert( p.first );
+	      }
+	    }
+	    for ( const auto& s : out_set ){
+	      line += s + ",";
+	    }
+	    os << line;
+	  }
+	}
+	else {
+	  vector<string> parts = TiCC::split( s );
+	  for ( const auto& w : parts ){
+	    UnicodeString uWord = TiCC::UnicodeFromUTF8(w);
+	    myMblem.Classify( uWord );
+	    vector<pair<string,string> > res = myMblem.getResult();
+	    string line = w + "\t";
+	    for ( const auto& p : res ){
+	      line += p.first + "[" + p.second + "]/";
+	    }
+	    line.erase(line.length()-1);
+	    line += "\n";
+	    os << line;
 	  }
-	  line.erase(line.length()-1);
-	  cout << line << endl;
 	}
       }
-      cout << "<utt>" << endl << endl;
+      if ( !usewordlist ){
+	os << "<utt>\n\n";
+      }
     }
-    cout << endl;
+    os << "\n";
   }
+  os.flush();
   return;
 }
 
 
 int main(int argc, char *argv[]) {
   std::ios_base::sync_with_stdio(false);
-  cerr << "mblem " << VERSION << " (c) CLTS, ILK 2014 - 2018" << endl;
+  cerr << "mblem " << VERSION << " (c) CLTS, ILK 2014 - 2020" << endl;
   cerr << "CLST  - Centre for Language and Speech Technology,"
        << "Radboud University" << endl
        << "ILK   - Induction of Linguistic Knowledge Research Group,"
        << "Tilburg University" << endl;
-  TiCC::CL_Options Opts("c:t:hVd:", "version,notagger,notokenizer");
+  TiCC::CL_Options Opts("c:t:hVd:o:", "version,notagger,notokenizer,wordlist");
   try {
     Opts.init(argc, argv);
   }
@@ -217,21 +277,32 @@ int main(int argc, char *argv[]) {
   }
   cerr << "based on [" << Timbl::VersionName() << "]" << endl;
   cerr << "configdir: " << configDir << endl;
+
   if ( parse_args(Opts) ){
     if (  !init() ){
       cerr << "terminated." << endl;
       return EXIT_FAILURE;
     }
+    ostream *os;
+    if ( !output_name.empty() ){
+      os = new ofstream( output_name );
+    }
+    else {
+      os = &cout;
+    }
     for ( const auto& TestFileName : fileNames ){
       ifstream in(TestFileName);
       if ( in.good() ){
-	Test( in );
+	Test( in, *os );
       }
       else {
 	cerr << "unable to open: " << TestFileName << endl;
 	continue;
       }
     }
+    if ( os != &cout ){
+      delete os;
+    }
   }
   else {
     return EXIT_FAILURE;
