@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -50,6 +50,7 @@ using TiCC::operator<<;
 #define LOG *TiCC::Log(myLog)
 
 string toString( const Compound::Type& ct ){
+  /// return the string representation for the Compound::Type
   switch ( ct ){
   case Compound::Type::NN:
     return "NN";
@@ -112,6 +113,7 @@ string toString( const Compound::Type& ct ){
 }
 
 Compound::Type stringToCompound( const string& s ){
+  /// give the Compound::Type for this string
   if ( s == "NN" ){
     return Compound::Type::NN;
   }
@@ -202,11 +204,13 @@ Compound::Type stringToCompound( const string& s ){
 }
 
 ostream& operator<<( ostream& os, const Compound::Type& ct ){
+  /// output a CompoundType to a stream
   os << toString( ct );
   return os;
 }
 
 string toString( const Status& st ){
+  /// return the string representation so a Status value
   switch ( st ){
   case Status::INFO:
     return "info";
@@ -230,15 +234,24 @@ string toString( const Status& st ){
 }
 
 ostream& operator<<( ostream& os, const Status& st ){
+  /// output a Status to a stream
   os << toString( st );
   return os;
 }
 
-BracketLeaf::BracketLeaf( const RulePart& p, int flag, TiCC::LogStream& l ):
-  BaseBracket(p.ResultClass, p.RightHand, flag, l ),
+BracketLeaf::BracketLeaf( const RulePart& p,
+			  int debug_flag,
+			  TiCC::LogStream& l ):
+  BaseBracket(p.ResultClass, p.RightHand, debug_flag, l ),
   glue(false),
   morph(p.morpheme )
 {
+  /// create a BracketLeaf object from a RulePart
+  /*!
+    \param p A Rulepart to create from
+    \param debug_flag the debug value
+    \param l a LogStream for messages
+  */
   ifpos = -1;
   if ( !p.inflect.empty() ){
     inflect = p.inflect;
@@ -282,28 +295,57 @@ BracketLeaf::BracketLeaf( const RulePart& p, int flag, TiCC::LogStream& l ):
 }
 
 BracketLeaf::BracketLeaf( CLEX::Type t,
-			  const icu::UnicodeString& us,
-			  int flag,
+			  const icu::UnicodeString& morpheme,
+			  int debug_flag,
 			  TiCC::LogStream& l ):
-  BaseBracket( t, vector<CLEX::Type>(), flag, l ),
-  morph( us )
+  BaseBracket( t, vector<CLEX::Type>(), debug_flag, l ),
+  morph( morpheme )
 {
+  /// create a BracketLeaf object from a CLEX::Type and a morpheme
+  /*!
+    \param t A CLEX::Type
+    \param morpheme the (Unicode) morpheme
+    \param debug_flag the debug value
+    \param l a LogStream for messages
+  */
   ifpos = -1;
   orig = toString( t );
   _status = Status::STEM;
 }
 
+BracketLeaf *BracketLeaf::clone() const{
+  /// make a copy of the BracketLeaf
+  return new BracketLeaf( *this );
+}
+
 BracketNest::BracketNest( CLEX::Type t,
 			  Compound::Type c,
-			  int flag,
+			  int debug_flag,
 			  TiCC::LogStream& l ):
-  BaseBracket( t, flag, l ),
+  BaseBracket( t, debug_flag, l ),
   _compound( c )
 {
+  /// create a BracketNest object from a CLEX::Type and a CompoundType
+  /*!
+    \param t A CLEX::Type
+    \param c a Compound::Type
+    \param debug_flag the debug value
+    \param l a LogStream for messages
+  */
   _status = Status::COMPLEX;
 }
 
+BracketNest *BracketNest::clone() const {
+  /// make a deep copy of the Bracketnest
+  BracketNest *result = new BracketNest( *this );
+  for ( auto& it : result->parts ){
+    it = it->clone();
+  }
+  return result;
+}
+
 BaseBracket *BracketNest::append( BaseBracket *t ){
+  /// append a Bracket structure to this Nest
   parts.push_back( t );
   return this;
 }
@@ -315,6 +357,7 @@ BracketNest::~BracketNest(){
 }
 
 icu::UnicodeString BaseBracket::put( bool full ) const {
+  /// create a Unicode string representation for this object
   icu::UnicodeString result = "[err?]";
   if ( full ){
     icu::UnicodeString s = TiCC::UnicodeFromUTF8(toString(cls));
@@ -323,7 +366,16 @@ icu::UnicodeString BaseBracket::put( bool full ) const {
   return result;
 }
 
+icu::UnicodeString BaseBracket::pretty_put() const {
+  /// create a descriptive Unicode string representation for this object
+  icu::UnicodeString result = "[err?]";
+  icu::UnicodeString s = TiCC::UnicodeFromUTF8(CLEX::get_tDescr(cls));
+  result += s;
+  return result;
+}
+
 icu::UnicodeString BracketLeaf::put( bool full ) const {
+  /// create a Unicode string representation for this object
   icu::UnicodeString result;
   if ( !morph.isEmpty() ){
     result += "[";
@@ -347,15 +399,48 @@ icu::UnicodeString BracketLeaf::put( bool full ) const {
   return result;
 }
 
+icu::UnicodeString BracketLeaf::pretty_put() const {
+  /// create a descriptive Unicode string representation for this object
+  string result;
+  if ( !morph.isEmpty() ){
+    result += "[";
+    result += TiCC::UnicodeToUTF8(morph);
+    result += "]";
+  }
+  if ( glue ){
+    string::size_type pos = orig.find( "^" );
+    string tag;
+    tag += orig[pos+1];
+    result += CLEX::get_tDescr(CLEX::toCLEX(tag));
+  }
+  if ( status() != Status::PARTICIPLE
+       && status() != Status::PARTICLE
+       && status() != Status::DERIVATIONAL
+       &&  status() != Status::FAILED
+       && cls != CLEX::UNASS
+       && cls != CLEX::NEUTRAL ){
+    //    LOG << this << " " << this->status() << endl;
+    string s = CLEX::get_tDescr(cls);
+    if ( s != "/" ){
+      result += s;
+    }
+  }
+  for ( const auto& i : inflect ){
+    string id = CLEX::get_iDescr(i);
+    if ( !id.empty() ){
+      result += "/" + id;
+    }
+  }
+  return TiCC::UnicodeFromUTF8(result);
+}
+
 icu::UnicodeString BracketNest::put( bool full ) const {
+  /// create a Unicode string representation for this object
   icu::UnicodeString result = "[ ";
   for ( auto const& it : parts ){
     icu::UnicodeString m = it->put( full );
     if ( !m.isEmpty() ){
       result += m + " ";
-      // if (&it != &parts.back() ){
-      // 	result += " ";
-      // }
     }
   }
   result += "]";
@@ -370,12 +455,34 @@ icu::UnicodeString BracketNest::put( bool full ) const {
   return result;
 }
 
+icu::UnicodeString BracketNest::pretty_put( ) const {
+  /// create a descriptive Unicode string representation for this object
+  icu::UnicodeString result;
+  int cnt = 0;
+  for ( auto const& it : parts ){
+    icu::UnicodeString m = it->pretty_put();
+    if ( m[0] == '[' ){
+      ++cnt;
+    }
+    result += m;
+  }
+  if ( cnt > 1 ){
+    result = "[" + result + "]";
+    if ( cls != CLEX::UNASS && cls != CLEX::NEUTRAL ){
+      result += TiCC::UnicodeFromUTF8(CLEX::get_tDescr(cls));
+    }
+  }
+  return result;
+}
+
 ostream& operator<< ( ostream& os, const BaseBracket& c ){
+  /// output a BaseBracket to a stream
   os << c.put();
   return os;
 }
 
 ostream& operator<< ( ostream& os, const BaseBracket *c ){
+  /// output a BaseBracket to a stream
   if ( c ){
     os << c->put();
   }
@@ -385,20 +492,16 @@ ostream& operator<< ( ostream& os, const BaseBracket *c ){
   return os;
 }
 
-void prettyP( ostream& os, const list<BaseBracket*>& v ){
-  os << "[";
-  for ( auto const& it : v ){
-    os << it;
-    if ( &it != &v.back() ){
-      os << " ";
-    }
-  }
-  os << "]";
-}
-
 bool BracketNest::testMatch( list<BaseBracket*>& result,
 			     const list<BaseBracket*>::iterator& rpos,
 			     list<BaseBracket*>::iterator& bpos ){
+  /// test if the rule matches at a certain position
+  /*!
+    \param result the current result. A new match will be appended
+    \param rpos the position in the rules list we are at.
+    \param bpos output parameter to return the END postion of the match
+    \return true if it matches
+  */
   if ( debugFlag > 5 ){
     LOG << "test MATCH: rpos= " << *rpos << endl;
   }
@@ -455,6 +558,11 @@ bool BracketNest::testMatch( list<BaseBracket*>& result,
 }
 
 Compound::Type construct( const vector<CLEX::Type>& tags ){
+  /// construct a Compound::Type given a list of CLEX tags
+  /*!
+    \param tags a list of CLEX::Types
+    \return a Compound::Type
+  */
   string s;
   for ( const auto& t : tags ){
     s += toString( t );
@@ -468,14 +576,24 @@ Compound::Type construct( const vector<CLEX::Type>& tags ){
 }
 
 Compound::Type construct( const CLEX::Type tag1, const CLEX::Type tag2 ){
+  /// construct a Compound::Type given two CLEX tags
+  /*!
+    \param tag1 a CLEX::Type
+    \param tag2 a CLEX::Type
+    \return a Compound::Type
+  */
   vector<CLEX::Type> v;
   v.push_back( tag1 );
   v.push_back( tag2 );
   return construct( v );
 }
 
-bool TEST = 1;
 Compound::Type BracketNest::getCompoundType(){
+  /// extract the Compound::Type
+  /*!
+    This function uses a lot of heuristics to determine get Compound::Type
+    given the elements in the BracketNest
+  */
   if ( debugFlag > 5 ){
     LOG << "get compoundType: " << this << endl;
     LOG << "#parts: " << parts.size() << endl;
@@ -566,8 +684,8 @@ Compound::Type BracketNest::getCompoundType(){
 	 && st1 != Status::PARTICIPLE ){
       switch ( tag1 ){
       case CLEX::N:
-	if ( st2 == Status::STEM && tag2 == CLEX::N
-	     && st3 == Status::STEM && tag3 == CLEX::N ){
+	if ( ( st2 == Status::STEM || st2 == Status::COMPLEX ) && tag2 == CLEX::N
+	     && (st3 == Status::STEM || st3 == Status::COMPLEX ) && tag3 == CLEX::N ){
 	  compound = Compound::Type::NNN;
 	}
 	else if ( st1 != Status::DERIVATIONAL && st2 == Status::STEM &&
@@ -685,6 +803,10 @@ Compound::Type BracketNest::getCompoundType(){
 }
 
 folia::Morpheme *BracketLeaf::createMorpheme( folia::Document *doc ) const {
+  /// use the data in the Leaf to create a folia::Morpheme node
+  /*!
+    \param doc The FoLiA Document context
+  */
   string desc;
   int cnt = 0;
   return createMorpheme( doc, desc, cnt );
@@ -693,6 +815,12 @@ folia::Morpheme *BracketLeaf::createMorpheme( folia::Document *doc ) const {
 folia::Morpheme *BracketLeaf::createMorpheme( folia::Document *doc,
 					      string& desc,
 					      int& cnt ) const {
+  /// use the data in the Leaf to create a folia::Morpheme node
+  /*!
+    \param doc The FoLiA Document context
+    \param desc a decriptien note to add
+    \param cnt a counter for the number of handled morphemes
+  */
   folia::Morpheme *result = 0;
   desc.clear();
   string::size_type pos = orig.find( "^" );
@@ -727,6 +855,20 @@ folia::Morpheme *BracketLeaf::createMorpheme( folia::Document *doc,
     else {
       args["class"] = toString( tag() );
       desc = "[" + out + "]" + CLEX::get_tDescr( tag() ); // spread the word upwards!
+      folia::KWargs fargs;
+      fargs["subset"] = "structure";
+      if ( tag() == CLEX::SPEC
+	   || tag() == CLEX::LET ){
+	fargs["class"] = "[" + out + "]";
+      }
+      else {
+	fargs["class"] = desc;
+      }
+#pragma omp critical (foliaupdate)
+      {
+	folia::Feature *feat = new folia::Feature( fargs );
+	result->append( feat );
+      }
     }
 #pragma omp critical (foliaupdate)
     {
@@ -832,13 +974,6 @@ folia::Morpheme *BracketLeaf::createMorpheme( folia::Document *doc,
       folia::Feature *feat = new folia::Feature( args );
       result->append( feat );
     }
-    args.clear();
-//     args["set"] = Mbma::clex_tagset;
-//     args["class"] = orig;
-// #pragma omp critical (foliaupdate)
-//     {
-//       result->addPosAnnotation( args );
-//     }
   }
   else if ( _status == Status::INFO ){
     folia::KWargs args;
@@ -870,6 +1005,10 @@ folia::Morpheme *BracketLeaf::createMorpheme( folia::Document *doc,
 }
 
 folia::Morpheme *BracketNest::createMorpheme( folia::Document *doc ) const {
+  /// use the data in the Leaf to create a folia::Morpheme node
+  /*!
+    \param doc The FoLiA Document context
+  */
   string desc;
   int cnt = 0;
   return createMorpheme( doc, desc, cnt );
@@ -878,6 +1017,12 @@ folia::Morpheme *BracketNest::createMorpheme( folia::Document *doc ) const {
 folia::Morpheme *BracketNest::createMorpheme( folia::Document *doc,
 					      string& desc,
 					      int& cnt ) const {
+  /// use the data in the Leaf to create a folia::Morpheme node
+  /*!
+    \param doc The FoLiA Document context
+    \param desc a decriptien note to add
+    \param cnt a counter for the number of handled morphemes
+  */
   folia::Morpheme *result = 0;
   folia::KWargs args;
   args["class"] = "complex";
@@ -920,6 +1065,9 @@ folia::Morpheme *BracketNest::createMorpheme( folia::Document *doc,
   cnt = 1;
   args.clear();
   args["subset"] = "structure";
+  if ( desc.empty() ){
+    desc = "XYZ";
+  }
   args["class"]  = desc;
 #pragma omp critical (foliaupdate)
   {
@@ -954,6 +1102,12 @@ folia::Morpheme *BracketNest::createMorpheme( folia::Document *doc,
 
 list<BaseBracket*>::iterator BracketNest::resolveAffix( list<BaseBracket*>& result,
 							const list<BaseBracket*>::iterator& rpos ){
+  /// try to resolve an Affix rule
+  /*!
+    \param result the output, matches might replace part of it by a new Nest
+    \param rpos start posotion for this search
+    \return an iterator where the next resolving step should start
+  */
   if ( debugFlag > 5 ){
     LOG << "resolve affix" << endl;
   }
@@ -998,6 +1152,7 @@ list<BaseBracket*>::iterator BracketNest::resolveAffix( list<BaseBracket*>& resu
 }
 
 void BracketNest::resolveNouns( ){
+  /// check for adjacent Nouns and replace by a new Nest
   if ( debugFlag > 5 ){
     LOG << "resolve NOUNS in:" << this << endl;
   }
@@ -1042,6 +1197,7 @@ void BracketNest::resolveNouns( ){
 
 list<BaseBracket*>::iterator BracketNest::glue( list<BaseBracket*>& result,
 						const list<BaseBracket*>::iterator& rpos ){
+  /// apply a glue rule
   if ( debugFlag > 5 ){
     LOG << "glue " << endl;
     LOG << "result IN : " << result << endl;
@@ -1112,6 +1268,7 @@ list<BaseBracket*>::iterator BracketNest::glue( list<BaseBracket*>& result,
 
 
 void BracketNest::resolveGlue( ){
+  /// resolve all glue rules
   list<BaseBracket*>::iterator it = parts.begin();
   while ( it != parts.end() ){
     // search for glue rules
@@ -1128,6 +1285,7 @@ void BracketNest::resolveGlue( ){
 }
 
 void BracketNest::resolveLead( ){
+  /// resolve rules starting with *
   list<BaseBracket*>::iterator it = parts.begin();
   while ( it != parts.end() ){
     // search for rules with a * at the begin
@@ -1153,6 +1311,7 @@ void BracketNest::resolveLead( ){
 }
 
 void BracketNest::resolveTail(){
+  /// resolve rules ending with *
   list<BaseBracket *>::iterator it = parts.begin();
   while ( it != parts.end() ){
     // search for rules with a * at the end
@@ -1185,6 +1344,7 @@ void BracketNest::resolveTail(){
 }
 
 void BracketNest::resolveMiddle(){
+  /// resolve rules with a * NOT at begin or end
   list<BaseBracket*>::iterator it = parts.begin();
   while ( it != parts.end() ){
     // now search for other rules with a * in the middle
@@ -1212,7 +1372,7 @@ void BracketNest::resolveMiddle(){
 }
 
 void BracketNest::clearEmptyNodes(){
-  // remove all nodes that don't have a morpheme or an inlection
+  /// remove all nodes that don't have a morpheme or an inflection
   if ( debugFlag > 5 ){
     LOG << "clear emptyNodes: " << this << endl;
   }
@@ -1247,6 +1407,9 @@ void BracketNest::clearEmptyNodes(){
 }
 
 CLEX::Type BracketNest::getFinalTag() {
+  /// get the result tag for this rule
+  // It is the last tag in the list, except for 'P' tags
+  //
   // LOG << "get Final Tag from: " << this << endl;
   CLEX::Type result_cls = CLEX::UNASS;
   auto it = parts.rbegin();
@@ -1271,57 +1434,6 @@ CLEX::Type BracketNest::getFinalTag() {
 	break;
       }
     }
-#ifdef NO_WAY
-    else if ( !(*it)->inflection().empty()
-	      && (*it)->morpheme().isEmpty() ){
-      string inf = (*it)->inflection();
-      // it is an inflection tag
-      LOG << " inflection: >" << inf << "<" << endl;
-      // given the specific selections of certain inflections,
-      //    select a tag!
-      CLEX::Type new_tag = CLEX::UNASS;
-      for ( size_t i=0; i < inf.size(); ++i ){
-	new_tag = CLEX::select_tag( inf[i] );
-	if ( new_tag != CLEX::UNASS ){
-	  LOG << inf[i] << " selects " << new_tag << endl;
-	  break;
-	}
-      }
-      if ( new_tag != CLEX::UNASS ) {
-	// apply the change. Remember, the idea is that an inflection is
-	// far more certain of the tag of its predecessing morpheme than
-	// the morpheme itself.
-	// This is not always the case, but it works
-	//
-	// go back to the previous morpheme
-	auto pit = it;
-	for( ++pit; pit != parts.rend(); ++pit ){
-	  CLEX::Type old_tag = (*pit)->tag();
-	  LOG << "een terug is " << *pit << endl;
-	  if ( CLEX::isBasicClass( old_tag ) &&
-	       old_tag != CLEX::P ){
-	    // only nodes that can get inflected (and unanalysed too)
-	    // now see if we can replace this class for a better one
-	    if ( old_tag == CLEX::PN && new_tag == CLEX::N ){
-	      LOG << "Don't replace PN by N" << endl;
-	    }
-	    else {
-	      LOG << " replace " << old_tag
-		  << " by " << new_tag << endl;
-	      (*pit)->setTag( new_tag );
-	      cls = new_tag;
-	    }
-	    return new_tag;
-	  }
-	}
-      }
-      else {
-	// this realy shouldn't happen. probably an error in the data!?
-	LOG << "inflection: " << inf
-	    << " Problem: DOESN'T select a tag" << endl;
-      }
-    }
-#endif
     ++it;
   }
   //  LOG << "final tag = " << result_cls << endl;
