@@ -1,6 +1,6 @@
 /* ex: set tabstop=8 expandtab: */
 /*
-  Copyright (c) 2006 - 2018
+  Copyright (c) 2006 - 2020
   CLST  - Radboud University
   ILK   - Tilburg University
 
@@ -40,99 +40,90 @@
 #include <map>
 
 #include "timbl/TimblAPI.h"
+#include "ticcutils/PrettyPrint.h"
 
 #include "frog/Frog-util.h" // defines etc.
 
 using namespace std;
 
-#define LOG *TiCC::Log(mwuLog)
+#define LOG *TiCC::Log(errLog)
+#define DBG *TiCC::Log(dbgLog)
 
-mwuAna::mwuAna( folia::Word *fwrd, const string& txt, const string& glue_tag ){
-  spec = false;
-  word = txt;
-  string tag;
-#pragma omp critical (foliaupdate)
-  {
-    tag = fwrd->annotation<folia::PosAnnotation>()->cls();
-  }
-  spec = ( tag == glue_tag );
-  fwords.push_back( fwrd );
+mwuAna::mwuAna( const string& wrd,
+		bool glue_tag,
+		size_t index ){
+  /// create a mwu Analysis record
+  /*!
+    \param wrd The text of the word
+    \param glue_tag was a \e 'glue' tag detected?
+    \param index The (initial) position in the sentence
+   */
+  word = wrd;
+  spec = glue_tag;
+  mwu_start = mwu_end = index;
 }
 
 void mwuAna::merge( const mwuAna *add ){
-  fwords.push_back( add->fwords[0] );
+  /// merge two mwuAna records
+  /*
+    \param add the mwuAna to merge.
+
+    in fact we only take over the endpostion of add the other fields are
+    of no interest more, as the words are already resolved
+   */
+  mwu_end = add->mwu_end;
   delete add;
 }
 
-folia::EntitiesLayer *mwuAna::addEntity( const string& tagset,
-					 const string& textclass,
-					 folia::Sentence *sent,
-					 folia::EntitiesLayer *el ){
-  if ( fwords.size() > 1 ){
-    if ( el == 0 ){
-#pragma omp critical (foliaupdate)
-      {
-	folia::KWargs args;
-	args["generate_id"] = sent->id();
-	args["set"] = tagset;
-	el = new folia::EntitiesLayer( args, sent->doc() );
-	sent->append( el );
-      }
-    }
-    folia::KWargs args;
-    args["set"] = tagset;
-    args["generate_id"] = el->id();
-    if ( textclass != "current" ){
-      args["textclass"] = textclass;
-    }
-    folia::Entity *e=0;
-#pragma omp critical (foliaupdate)
-    {
-      e = new folia::Entity( args, el->doc() );
-      el->append( e );
-    }
-    for ( const auto& fw : fwords ){
-#pragma omp critical (foliaupdate)
-      {
-	e->append( fw );
-      }
-    }
-  }
-  return el;
-}
+Mwu::Mwu( TiCC::LogStream *err_log, TiCC::LogStream *dbg_log ){
+  /// create a Mwu record (UNINITIALIZED yet)
+  /*!
+    \param err_log The LogStream for errors
+    \param dbg_log The LogStream for debugging
 
-Mwu::Mwu( TiCC::LogStream * logstream ){
-  mwuLog = new TiCC::LogStream( logstream, "mwu-" );
+    a call to Mwu::init() is needed to be able to use the Mwu classifier
+   */
+  errLog = new TiCC::LogStream( err_log, "mwu-" );
+  dbgLog = new TiCC::LogStream( dbg_log, "mwu-" );
   filter = 0;
 }
 
 Mwu::~Mwu(){
+  /// erase an Mwu record
   reset();
-  delete mwuLog;
+  delete errLog;
+  delete dbgLog;
   delete filter;
 }
 
 void Mwu::reset(){
+  /// clear an Mwu record
   for ( const auto& it : mWords ){
     delete it;
   }
   mWords.clear();
 }
 
-void Mwu::add( folia::Word *word ){
-  icu::UnicodeString tmp;
-#pragma omp critical (foliaupdate)
-  {
-    tmp = word->text( textclass );
-  }
-  if ( filter )
+void Mwu::add( frog_record& fd ){
+  /// add a mwuAna record to this Mwu
+  /*!
+    \param fd The frog_data structure with the information to use
+   */
+  icu::UnicodeString tmp = TiCC::UnicodeFromUTF8(fd.word);
+  if ( filter ){
     tmp = filter->filter( tmp );
+  }
   string txt = TiCC::UnicodeToUTF8( tmp );
-  mWords.push_back( new mwuAna( word, txt, glue_tag ) );
+  bool glue = ( fd.tag == glue_tag );
+  size_t index = mWords.size();
+  mWords.push_back( new mwuAna( txt, glue, index ) );
 }
 
-
 bool Mwu::read_mwus( const string& fname) {
+  /// fill our table with MWU's
+  /*!
+    \param fname the file to reaf from
+   */
   LOG << "read mwus " + fname << endl;
   ifstream mwufile(fname, ios::in);
   if(mwufile.bad()){
@@ -156,6 +147,10 @@ bool Mwu::read_mwus( const string& fname) {
 }
 
 bool Mwu::init( const TiCC::Configuration& config ) {
+  /// initialize the Mwu using a Configuration structure
+  /*!
+    \param config the configuration to use
+   */
   LOG << "initiating mwuChunker..." << endl;
   debug = 0;
   string val = config.lookUp( "debug", "mwu" );
@@ -177,10 +172,10 @@ bool Mwu::init( const TiCC::Configuration& config ) {
   }
   val = config.lookUp( "version", "mwu" );
   if ( val.empty() ){
-    version = "1.0";
+    _version = "1.0";
   }
   else {
-    version = val;
+    _version = val;
   }
   val = config.lookUp( "set", "mwu" );
   if ( val.empty() ){
@@ -190,8 +185,9 @@ bool Mwu::init( const TiCC::Configuration& config ) {
     mwu_tagset = val;
   }
   string charFile = config.lookUp( "char_filter_file", "tagger" );
-  if ( charFile.empty() )
+  if ( charFile.empty() ){
     charFile = config.lookUp( "char_filter_file" );
+  }
   if ( !charFile.empty() ){
     charFile = prefix( config.configDir(), charFile );
     filter = new TiCC::UniFilter();
@@ -216,42 +212,71 @@ bool Mwu::init( const TiCC::Configuration& config ) {
   return true;
 }
 
+using TiCC::operator<<;
+
 ostream &operator<<( ostream& os, const Mwu& mwu ){
-  for ( size_t i = 0; i < mwu.mWords.size(); ++i )
-    os << i+1 << "\t" << mwu.mWords[i]->getWord() << endl;
+  /// output an Mwu (debugging only)
+  for ( size_t i = 0; i < mwu.mWords.size(); ++i ){
+    os << i+1 << "\t" << mwu.mWords[i]->getWord()
+       <<  "(" << mwu.mWords[i]->mwu_start << "," << mwu.mWords[i]->mwu_end
+       << ")" << endl;
+  }
   return os;
 }
 
-void Mwu::addDeclaration( folia::Document& doc ) const {
-  doc.declare( folia::AnnotationType::ENTITY,
-	       mwu_tagset,
-	       "annotator='frog-mwu-" + version
-	       + "', annotatortype='auto', datetime='" + getTime() + "'");
+void Mwu::add_provenance( folia::Document& doc,
+			    folia::processor *main ) const {
+  /// add provenance information for the tokenizer. (FoLiA output only)
+  /*!
+    \param doc the FoLiA document to add to
+    \param main the processor to use (presumably the Frog processor)
+  */
+  string _label = "mwu";
+  if ( !main ){
+    throw logic_error( "mwu::add_provenance() without arguments." );
+  }
+  folia::KWargs args;
+  args["name"] = _label;
+  args["generate_id"] = "auto()";
+  args["version"] = _version;
+  args["begindatetime"] = "now()";
+  folia::processor *proc = doc.add_processor( args, main );
+  args.clear();
+  args["processor"] = proc->id();
+  doc.declare( folia::AnnotationType::ENTITY, mwu_tagset, args );
 }
 
-void Mwu::Classify( const vector<folia::Word*>& words ){
-  if ( words.empty() ){
-    return;
-  }
+void Mwu::Classify( frog_data& sent ){
+  /// run the Mwu classifier on e sentence in frog_data format
+  /*!
+    \param sent a frog_data structure with unresolved MWU's
+   */
   reset();
-  for ( const auto& word : words ){
+  // setup the Mwu
+  for ( auto& word : sent.units ){
     add( word );
   }
+  // collect mwu's
   Classify();
-  folia::EntitiesLayer *el = 0;
-  folia::Sentence *sent;
-#pragma omp critical (foliaupdate)
-  {
-    sent = words[0]->sentence();
-  }
+  //  take over the mwu postions in the frog_data
   for ( const auto& mword : mWords ){
-    el = mword->addEntity( mwu_tagset, textclass, sent, el );
+    if ( mword->mwu_start != mword->mwu_end ){
+      sent.mwus[mword->mwu_start] = mword->mwu_end;
+    }
   }
+  sent.resolve_mwus();
 }
 
 void Mwu::Classify(){
-  if ( debug ) {
-    LOG << "Starting mwu Classify" << endl;
+  /// examine the Mwu's internal mwuAna nodes to determine the spans of
+  /// all mwu's found
+  /*!
+    First we collect all 'glue tags' into MWU's
+
+    Second we do a lookup of the words in our table and keep the longest match
+   */
+  if ( debug > 1 ) {
+    DBG << "Starting mwu Classify" << endl;
   }
   mymap2::iterator best_match;
   size_t matchLength = 0;
@@ -273,36 +298,36 @@ void Mwu::Classify(){
   size_t i;
   for ( i = 0; i < max; i++) {
     string word = mWords[i]->getWord();
-    if ( debug ){
-      LOG << "checking word[" << i <<"]: " << word << endl;
+    if ( debug > 1 ){
+      DBG << "checking word[" << i <<"]: " << word << endl;
     }
     const auto matches = MWUs.equal_range(word);
     if ( matches.first != MWUs.end() ) {
       //match
       auto current_match = matches.first;
-      if (  debug  ) {
-	LOG << "MWU: match found for " << word << endl;
+      if (  debug > 1 ) {
+	DBG << "MWU: match found for " << word << endl;
       }
       while( current_match != matches.second
 	     && current_match != MWUs.end() ){
 	vector<string> match = current_match->second;
 	size_t max_match = match.size();
 	size_t j = 0;
-	if ( debug ){
-	  LOG << "checking " << max_match << " matches:" << endl;
+	if ( debug > 1 ){
+	  DBG << "checking " << max_match << " matches:" << endl;
 	}
 	for (; i + j + 1 < max && j < max_match; j++) {
 	  if ( match[j] != mWords[i+j+1]->getWord() ) {
-	    if ( debug ){
-	      LOG << "match " << j <<" (" << match[j]
+	    if ( debug > 1){
+	      DBG << "match " << j <<" (" << match[j]
 			   << ") doesn't match with word " << i+ j + 1
 			   << " (" << mWords[i+j + 1]->getWord() <<")" << endl;
 	    }
 	    // mismatch in jth word of current mwu
 	    break;
 	  }
-	  else if ( debug ){
-	    LOG << " matched " <<  mWords[i+j+1]->getWord()
+	  else if ( debug > 1 ){
+	    DBG << " matched " <<  mWords[i+j+1]->getWord()
 			 << " j=" << j << endl;
 	  }
 
@@ -314,44 +339,82 @@ void Mwu::Classify(){
 	}
 	++current_match;
       } // while
-      if( debug ){
+      if( debug > 1){
 	if (matchLength >0 ) {
-	  LOG << "MWU: found match starting with " << (*best_match).first << endl;
+	  DBG << "MWU: found match starting with " << (*best_match).first << endl;
 	}
 	else {
-	  LOG <<"MWU: no match" << endl;
+	  DBG <<"MWU: no match" << endl;
 	}
       }
-      // we found a matching mwu, break out of loop thru sentence,
+      // we found a matching mwu, break out of the loop
       // do useful stuff, and recurse to find more mwus
       if ( matchLength > 0 ){
 	break;
       }
     } //match found
     else {
-      if( debug )
-	LOG <<"MWU:check: no match" << endl;
+      if( debug > 1 ) {
+	DBG <<"MWU:check: no match" << endl;
+      }
     }
   } //for (i < max)
   if (matchLength > 0 ) {
     //concat
-    if ( debug ){
-      LOG << "mwu found, processing" << endl;
+    if ( debug >1 ){
+      DBG << "mwu found, processing" << endl;
     }
     for ( size_t j = 1; j <= matchLength; ++j) {
-      if ( debug ){
-	LOG << "concat " << mWords[i+j]->getWord() << endl;
+      if ( debug > 1 ){
+	DBG << "concat " << mWords[i+j]->getWord() << endl;
       }
       mWords[i]->merge( mWords[i+j] );
     }
     vector<mwuAna*>::iterator anatmp1 = mWords.begin() + i;
     vector<mwuAna*>::iterator anatmp2 = ++anatmp1 + matchLength;
     mWords.erase(anatmp1, anatmp2);
-    if ( debug ){
-      LOG << "tussenstand:" << endl;
-      LOG << *this << endl;
+    if ( debug > 1){
+      DBG << "tussenstand:" << endl;
+      DBG << *this << endl;
     }
     Classify( );
   } //if (matchLength)
   return;
 } // //Classify
+
+void Mwu::add_result( const frog_data& fd,
+		      const vector<folia::Word*>& wv ) const {
+  /// add the mwu's in \e fd as Entities to the parent of the \e wv list of Word
+  /*!
+    \param fd The tagged results
+    \param wv The folia:Word vector
+  */
+  folia::Sentence *s = wv[0]->sentence();
+  folia::KWargs args;
+  if ( !s->id().empty() ){
+    args["generate_id"] = s->id();
+  }
+  args["set"] = getTagset();
+  folia::EntitiesLayer *el;
+#pragma omp critical (foliaupdate)
+  {
+    el = new folia::EntitiesLayer( args, s->doc() );
+    s->append( el );
+  }
+  for ( const auto& mwu : fd.mwus ){
+    if ( !el->id().empty() ){
+      args["generate_id"] = el->id();
+    }
+    if ( textclass != "current" ){
+      args["textclass"] = textclass;
+    }
+#pragma omp critical (foliaupdate)
+    {
+      folia::Entity *e = new folia::Entity( args, s->doc() );
+      el->append( e );
+      for ( size_t pos = mwu.first; pos <= mwu.second; ++pos ){
+	e->append( wv[pos] );
+      }
+    }
+  }
+}
